from datetime import datetime

from six.moves.urllib.parse import parse_qs, urlparse


from prometeo import exceptions
from prometeo.banking import exceptions as banking_exceptions
from prometeo.banking.client import Account, CreditCard
from prometeo.banking.models import Account as AccountModel
from prometeo.banking.models import CreditCard as CreditCardModel
from tests.base_test_case import BaseTestCase
import respx


class TestClient(BaseTestCase):
    @respx.mock
    def test_login_success(self):
        self.mock_post_request(
            respx,
            "/login/",
            json={
                "status": "logged_in",
                "key": "123456",
            },
        )
        session = self.client.banking.new_session()
        session.login(
            provider="test_provider",
            username="test_username",
            password="test_password",
        )
        self.assertEqual("logged_in", session.get_status())
        self.assertEqual("123456", session.get_session_key())

    @respx.mock
    def test_login_wrong_credentials(self):
        self.mock_post_request(
            respx,
            "/login/",
            status_code=403,
            json={
                "status": "wrong_credentials",
                "message": "wrong credentials",
            },
        )
        with self.assertRaises(exceptions.WrongCredentialsError):
            session = self.client.banking.new_session()
            session.login(
                provider="test_provider",
                username="test_username",
                password="test_password",
            )

    @respx.mock
    def test_generic_login_error(self):
        self.mock_post_request(
            respx,
            "/login/",
            status_code=200,
            json={
                "status": "error",
                "message": "An error has ocurred.",
            },
        )
        with self.assertRaises(banking_exceptions.BankingClientError):
            session = self.client.banking.new_session()
            session.login(
                provider="test_provider",
                username="test_username",
                password="test_password",
            )

    @respx.mock
    def test_login_select_client(self):
        self.mock_post_request(
            respx,
            "/login/",
            json={
                "status": "select_client",
                "key": "123456",
            },
        )
        self.mock_get_request(
            respx,
            "/client/",
            json={
                "status": "success",
                "clients": {"0": "First Client", "1": "Second Client"},
            },
        )
        self.mock_get_request(respx, "/client/1/", json={"status": "success"})
        session = self.client.banking.new_session()
        session.login(
            provider="test_provider",
            username="test_username",
            password="test_password",
        )
        self.assertEqual("select_client", session.get_status())
        self.assertEqual("123456", session.get_session_key())

        clients = session.get_clients()
        history = respx.calls
        self.assertEqual("/client/", history[-1].request.url.path)

        client = [client for client in clients if client.id == "1"][0]
        session.select_client(client)
        self.assertEqual("/client/1/", history[-1].request.url.path)

    @respx.mock
    def test_login_interactive(self):
        session_key = "123456"
        personal_question = "¿Cuántos baños tenia la casa de mis padres?"
        self.mock_post_request(
            respx,
            "/login/",
            json={
                "status": "interaction_required",
                "context": personal_question,
                "field": "personal_question",
                "key": session_key,
            },
        )
        session = self.client.banking.new_session()
        session.login(
            provider="test_provider",
            username="test_username",
            password="test_password",
        )
        self.assertEqual("interaction_required", session.get_status())
        self.assertEqual(session_key, session.get_session_key())
        self.assertEqual(personal_question, session.get_interactive_context())

        self.mock_post_request(
            respx,
            "/login/",
            json={
                "status": "logged_in",
            },
        )
        challenge_answer = "test answer"
        session.finish_login(
            "test_provider", "test_username", "test_password", challenge_answer
        )
        last_request = respx.calls.last.request
        request_body = parse_qs(str(last_request.content))
        self.assertIn(
            challenge_answer,
            request_body["personal_question"][0],
        )
        self.assertEqual(last_request.headers["X-Session-Key"], session_key)

    @respx.mock
    def test_get_accounts(self):
        self.mock_get_request(
            respx,
            "/account/",
            json={
                "accounts": [
                    {
                        "balance": 1234.95,
                        "branch": "02 - 18 De Julio",
                        "currency": "UYU",
                        "id": "hash1",
                        "name": "Cuenta total",
                        "number": "001234567890",
                    },
                    {
                        "balance": 53.96,
                        "branch": "61 - Ciudad Vieja",
                        "currency": "USD",
                        "id": "hash4",
                        "name": "Caja De Ahorro Atm",
                        "number": "004327567890",
                    },
                ],
                "status": "success",
            },
        )
        session_key = "test_session_key"
        session = self.client.banking.get_session(session_key)
        accounts = session.get_accounts()
        last_request = respx.calls.last.request
        self.assertEqual("/account/", str(last_request.url.path))
        self.assertEqual(session_key, last_request.headers["X-Session-Key"])
        self.assertEqual(2, len(accounts))
        self.assertEqual("Cuenta total", accounts[0].name)
        self.assertEqual("Caja De Ahorro Atm", accounts[1].name)

    @respx.mock
    def test_get_movements(self):
        self.mock_get_request(
            respx,
            "/movement/",
            json={
                "movements": [
                    {
                        "credit": "",
                        "date": "12/01/2019",
                        "debit": 3500,
                        "detail": "RETIRO EFECTIVO CAJERO AUTOMATICO J.C. ",
                        "id": "-890185180",
                        "reference": "000000005084",
                        "extra_data": None,
                    },
                    {
                        "credit": "",
                        "date": "05/07/2019",
                        "debit": 16000,
                        "detail": "RETIRO EFECTIVO CAJERO AUTOMATICO J.H Y ",
                        "id": 1024917397,
                        "reference": "000000002931",
                        "extra_data": None,
                    },
                ],
                "status": "success",
            },
        )
        session_key = "test_session_key"
        account_number = "001234567"
        currency_code = "USD"
        date_start = datetime(2019, 1, 1)
        date_end = datetime(2019, 12, 1)
        account = Account(
            self.client.banking,
            session_key,
            AccountModel(
                id="12345",
                name="Cuenta total",
                number="001234567",
                branch="02 - 18 De Julio",
                currency="USD",
                balance=1234.95,
            ),
        )
        movements = account.get_movements(date_start, date_end)
        last_request = respx.calls.last.request
        qs = parse_qs(urlparse(str(last_request.url)).query)
        self.assertEqual("/movement/", last_request.url.path)
        self.assertEqual(session_key, last_request.headers["X-Session-Key"])
        self.assertEqual(account_number, qs["account"][0])
        self.assertEqual(currency_code, qs["currency"][0])
        self.assertEqual("01/01/2019", qs["date_start"][0])
        self.assertEqual("01/12/2019", qs["date_end"][0])
        self.assertEqual(2, len(movements))
        self.assertEqual(datetime(2019, 1, 12), movements[0].date)
        self.assertEqual(datetime(2019, 7, 5), movements[1].date)

    @respx.mock
    def test_get_credit_cards(self):
        self.mock_get_request(
            respx,
            "/credit-card/",
            json={
                "credit_cards": [
                    {
                        "balance_dollar": 67.89,
                        "balance_local": 12345.42,
                        "close_date": "04/11/2019",
                        "due_date": "20/11/2019",
                        "id": "EcebS0rIZ5NdJNzS4XHmuOnHvSsI72TSA3j/8drzYYHtW1sQbpz5Hc",
                        "name": "Vi Int Sumaclub Plus",
                        "number": "770012345678",
                    }
                ],
                "status": "success",
            },
        )
        session_key = "test_session_key"
        session = self.client.banking.get_session(session_key)
        cards = session.get_credit_cards()
        last_request = respx.calls.last.request
        self.assertEqual("/credit-card/", last_request.url.path)
        self.assertEqual(session_key, last_request.headers["X-Session-Key"])
        self.assertEqual(1, len(cards))
        self.assertEqual("Vi Int Sumaclub Plus", cards[0].name)
        self.assertEqual("770012345678", cards[0].number)
        self.assertEqual(datetime(2019, 11, 20), cards[0].due_date)

    @respx.mock
    def test_get_credit_card_movements(self):
        self.mock_get_request(
            respx,
            "/credit-card/1234567/movements",
            json={
                "movements": [
                    {
                        "credit": "",
                        "date": "12/01/2019",
                        "debit": 3500,
                        "detail": "RETIRO EFECTIVO CAJERO AUTOMATICO J.C. ",
                        "id": "-890185180",
                        "reference": "000000005084",
                    },
                    {
                        "credit": "",
                        "date": "05/07/2019",
                        "debit": 16000,
                        "detail": "RETIRO EFECTIVO CAJERO AUTOMATICO J.H Y ",
                        "id": "1024917397",
                        "reference": "000000002931",
                    },
                ],
                "status": "success",
            },
        )
        session_key = "test_session_key"
        card_number = "1234567"
        currency_code = "USD"
        date_start = datetime(2019, 1, 1)
        date_end = datetime(2019, 12, 1)
        card = CreditCard(
            self.client.banking,
            session_key,
            CreditCardModel(
                id="12345",
                name="Cuenta total",
                number=card_number,
                close_date=datetime(2019, 11, 4),
                due_date=datetime(2019, 11, 20),
                balance_local=12345.42,
                balance_dollar=67.89,
            ),
        )
        movements = card.get_movements(currency_code, date_start, date_end)
        last_request = respx.calls.last.request
        qs = parse_qs(urlparse(str(last_request.url)).query)
        self.assertEqual("/credit-card/1234567/movements", last_request.url.path)
        self.assertEqual(session_key, last_request.headers["X-Session-Key"])
        self.assertEqual(currency_code, qs["currency"][0])
        self.assertEqual("01/01/2019", qs["date_start"][0])
        self.assertEqual("01/12/2019", qs["date_end"][0])
        self.assertEqual(2, len(movements))
        self.assertEqual(datetime(2019, 1, 12), movements[0].date)
        self.assertEqual(datetime(2019, 7, 5), movements[1].date)

    @respx.mock
    def test_get_providers(self):
        self.mock_get_request(
            respx,
            "/provider/",
            json={
                "providers": [
                    {"code": "test", "country": "UY", "name": "Test Provider"}
                ],
                "status": "success",
            },
        )
        session = self.client.banking.new_session()
        providers = session.get_providers()
        self.assertEqual(1, len(providers))
        self.assertEqual("test", providers[0].code)
        self.assertEqual("UY", providers[0].country)
        self.assertEqual("Test Provider", providers[0].name)

    @respx.mock
    def test_get_provider_detail(self):
        self.mock_get_request(
            respx,
            "/provider/test/",
            json={
                "provider": {
                    "auth_fields": [
                        {
                            "name": "username",
                            "type": "text",
                            "interactive": False,
                            "optional": False,
                            "label_es": "Usuario",
                            "label_en": "Username",
                        },
                        {
                            "name": "password",
                            "type": "password",
                            "interactive": False,
                            "optional": False,
                            "label_es": "Contraseña",
                            "label_en": "Password",
                        },
                        {
                            "name": "personal_question",
                            "type": "text",
                            "interactive": True,
                            "optional": False,
                            "label_es": "Pregunta personal",
                            "label_en": "Personal Question",
                        },
                        {
                            "name": "otp",
                            "type": "text",
                            "interactive": True,
                            "optional": False,
                            "label_es": "Token OTP",
                            "label_en": "OTP Token",
                        },
                    ],
                    "name": "test",
                    "aliases": ["alias_test"],
                    "country": "UY",
                    "endpoints_status": None,
                    "account_type": [
                        {
                            "name": "pers",
                            "label_es": "Cuenta Personal",
                            "label_en": "Personal Account",
                        },
                        {
                            "name": "corp",
                            "label_es": "Cuenta Corporativa",
                            "label_en": "Corporate Account",
                        },
                    ],
                    "logo": "https://providers.prometeoapi.com/logos/test.png",
                    "bank": {
                        "code": "test",
                        "name": "Banco Test Prometeo",
                        "logo": "https://providers.prometeoapi.com/logos/test.png",
                    },
                    "methods": {
                        "accounts": True,
                        "credit_cards": True,
                        "accounts_movements": True,
                        "credit_cards_movements": True,
                        "personal_info": True,
                        "transfers": True,
                        "enrollments": True,
                    },
                },
                "status": "success",
            },
        )
        session = self.client.banking.new_session()
        provider = session.get_provider_detail("test")
        self.assertEqual("UY", provider.country)
        self.assertEqual("test", provider.name)
        self.assertEqual(4, len(provider.auth_fields))

    @respx.mock
    def test_get_provider_detail_filtered_options(self):
        self.mock_get_request(
            respx,
            "/provider/santander_pers_uy/?key=type&value=UY",
            "provider_details_santander",
        )
        session = self.client.banking.new_session()
        provider = session.get_provider_detail("santander_pers_uy", "type", "UY")
        self.assertEqual("UY", provider.country)
        self.assertEqual("santander_pers_uy", provider.name)
        self.assertEqual(4, len(provider.auth_fields))

    @respx.mock
    def test_provider_doesnt_exist(self):
        self.mock_get_request(
            respx,
            "/provider/invalid/",
            status_code=404,
            json={
                "status": "provider_doesnt_exist",
            },
        )
        with self.assertRaises(exceptions.NotFoundError):
            session = self.client.banking.new_session()
            session.get_provider_detail("invalid")

    @respx.mock
    def test_logout(self):
        self.mock_get_request(respx, "/logout/", json={"status": "logged_out"})
        session_key = "test_session_key"
        session = self.client.banking.get_session(session_key)
        session.logout()
        last_request = respx.calls.last.request
        self.assertEqual("/logout/", last_request.url.path)
        self.assertEqual(session_key, last_request.headers["X-Session-Key"])

    @respx.mock
    def test_preprocess_transfer(self):
        self.mock_post_request(
            respx,
            "/transfer/preprocess",
            json={
                "result": {
                    "approved": True,
                    "authorization_devices": [
                        {"data": ["F-4", "B-2", "G-7"], "type": "cardCode"},
                        {"data": None, "type": "pin"},
                    ],
                    "message": None,
                    "request_id": "0b7d6b32d1be4c11bde21e7ddc08cc36",
                },
                "status": "success",
            },
        )
        session_key = "test_session_key"
        origin_account = "002206345988"
        destination_institution = "0"
        destination_account = "001002363321"
        currency = "UYU"
        amount = "1.3"
        concept = "descripcion de transferencia"
        destination_owner_name = "John Doe"
        branch = "62"
        session = self.client.banking.get_session(session_key)
        preprocess = session.preprocess_transfer(
            origin_account,
            destination_institution,
            destination_account,
            currency,
            amount,
            concept,
            destination_owner_name,
            branch,
        )
        self.assertEqual(True, preprocess.approved)
        self.assertEqual(3, len(preprocess.authorization_devices[0].data))
        self.assertEqual("cardCode", preprocess.authorization_devices[0].type)
        self.assertEqual("0b7d6b32d1be4c11bde21e7ddc08cc36", preprocess.request_id)
        self.assertEqual(None, preprocess.message)

    @respx.mock
    def test_confirm_transfer(self):
        self.mock_post_request(
            respx,
            "/transfer/confirm",
            json={
                "status": "success",
                "transfer": {
                    "message": "Transferencia confirmada con exito",
                    "success": True,
                },
            },
        )
        session_key = "test_session_key"
        request_id = "0b7d6b32d1be4c11bde21e7ddc08cc36"
        authorization_type = "cardCode"
        authorization_data = "1, 2, 3"
        session = self.client.banking.get_session(session_key)
        confirmation = session.confirm_transfer(
            request_id, authorization_type, authorization_data
        )
        self.assertEqual(True, confirmation.success)
        self.assertNotEqual(None, confirmation.message)

    @respx.mock
    def test_list_transfer_institutions(self):
        self.mock_get_request(
            respx,
            "/transfer/destinations",
            json={
                "destinations": [
                    {"id": 0, "name": "SANTANDER"},
                    {"id": 1, "name": "B.R.O.U."},
                    {"id": 91, "name": "B.H.U."},
                    {"id": 110, "name": "BANDES"},
                ],
                "status": "success",
            },
        )
        session_key = "test_session_key"
        session = self.client.banking.get_session(session_key)
        institutions = session.list_transfer_institutions()
        self.assertNotEqual(0, len(institutions))

    @respx.mock
    def test_invalid_session_key(self):
        self.mock_get_request(
            respx, "/account/", json={"message": "Invalid key", "status": "error"}
        )
        with self.assertRaises(exceptions.InvalidSessionKeyError):
            session = self.client.banking.new_session()
            session.get_accounts()

    @respx.mock
    def test_generic_client_error(self):
        self.mock_get_request(
            respx,
            "/account/",
            json={"message": "Some generic error", "status": "error"},
        )
        with self.assertRaises(banking_exceptions.BankingClientError):
            session = self.client.banking.new_session()
            session.get_accounts()



from datetime import datetime
import httpx
import six
from six.moves.urllib.parse import parse_qs
import respx

from prometeo.dian import (
    DocumentType,
    Periodicity,
    QuarterlyPeriod,
    NumerationType,
    MonthlyPeriod,
)
from tests.base_test_case import BaseTestCase


class TestClient(BaseTestCase):
    def setUp(self):
        super(TestClient, self).setUp()
        self.session_key = "123456"
        self.pdf_content = "pdf content"

    async def assert_pdf_downloaded(self, response):
        self.assertEqual(
            six.ensure_binary(self.pdf_content), await response.pdf.get_file()
        )

    def mock_pdf_download(self, mocker, url, fixture_name):
        response_data = self.load_json(fixture_name)
        key = [k for k in response_data.keys() if k != "status"][0]
        mocker.get(url, json=response_data)
        mocker.get(url).mock(
            return_value=httpx.Response(status_code=200, json=response_data)
        )
        mocker.get(response_data[key]["pdf_url"]).mock(
            return_value=httpx.Response(status_code=200, text=self.pdf_content)
        )

    @respx.mock
    async def test_login_company_success(self):
        self.mock_post_request(respx, "/login/", "successful_login")
        session = await self.client.dian.login(
            nit="098765",
            document_type=DocumentType.CEDULA_CIUDADANIA,
            document="12345",
            password="test_password",
        )
        self.assertEqual("logged_in", session.get_status())
        self.assertEqual(self.session_key, session.get_session_key())

        last_request = respx.calls.last.request
        request_body = parse_qs(str(last_request.content.decode("utf-8")))
        self.assertEqual("098765", request_body["nit"][0])
        self.assertEqual("13", request_body["document_type"][0])
        self.assertEqual("12345", request_body["document"][0])
        self.assertEqual("test_password", request_body["password"][0])

    @respx.mock
    async def test_login_person_success(self):
        self.mock_post_request(respx, "/login/", "successful_login")
        session = await self.client.dian.login(
            document_type=DocumentType.CEDULA_CIUDADANIA,
            document="12345",
            password="test_password",
        )
        self.assertEqual("logged_in", session.get_status())
        self.assertEqual("123456", session.get_session_key())

        last_request = respx.calls.last.request
        request_body = parse_qs(str(last_request.content.decode("utf-8")))
        self.assertNotIn("nit", request_body)
        self.assertEqual("13", request_body["document_type"][0])
        self.assertEqual("12345", request_body["document"][0])
        self.assertEqual("test_password", request_body["password"][0])

    @respx.mock
    async def test_get_company_info(self):
        self.mock_get_request(respx, "/company-info/", "company_info")
        info = await self.client.dian.get_company_info(self.session_key)

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual(info.reason, "Qualia Fintech SRL")
        self.assertEqual(info.location.country, "Uruguay")
        self.assertEqual(info.accountant.start_date, datetime(2017, 5, 1))

    @respx.mock
    async def test_download_company_info(self):
        self.mock_pdf_download(respx, "/company-info/", "company_info")
        info = await self.client.dian.get_company_info(self.session_key)
        await self.assert_pdf_downloaded(info)

    @respx.mock
    async def test_get_balances(self):
        self.mock_get_request(respx, "/balances/", "company_balances")
        balances = await self.client.dian.get_balances(self.session_key)
        last_request = respx.calls.last.request
        self.assertEqual(2, len(balances))
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])

    @respx.mock
    async def test_get_rent_declaration(self):
        self.mock_get_request(respx, "/rent/", "company_rent")
        rent = await self.client.dian.get_rent_declaration(self.session_key, 2019)

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual("2019", self.qs(last_request)["year"][0])
        self.assertEqual(rent.nit, "333222251")
        self.assertEqual(6, len(rent.fields))

    @respx.mock
    async def test_download_rent_declaration(self):
        self.mock_pdf_download(respx, "/rent/", "company_rent")
        rent = await self.client.dian.get_rent_declaration(self.session_key, 2019)
        await self.assert_pdf_downloaded(rent)

    @respx.mock
    async def test_get_vat_declaration(self):
        self.mock_get_request(respx, "/vat/", "vat_declaration")
        vat = await self.client.dian.get_vat_declaration(
            self.session_key,
            2019,
            Periodicity.QUARTERLY,
            QuarterlyPeriod.JANUARY_APRIL,
        )

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual("2019", self.qs(last_request)["year"][0])
        self.assertEqual("q", self.qs(last_request)["periodicity"][0])
        self.assertEqual("1", self.qs(last_request)["period"][0])
        self.assertEqual(vat.nit, "123332211")
        self.assertEqual(6, len(vat.fields))

    @respx.mock
    async def test_download_vat_declaration(self):
        self.mock_pdf_download(respx, "/vat/", "vat_declaration")
        vat = await self.client.dian.get_vat_declaration(
            self.session_key,
            2019,
            Periodicity.QUARTERLY,
            QuarterlyPeriod.JANUARY_APRIL,
        )
        await self.assert_pdf_downloaded(vat)

    @respx.mock
    async def test_get_numeration(self):
        self.mock_get_request(respx, "/numeration/", "numeration_authorization")
        numerations = await self.client.dian.get_numeration(
            self.session_key,
            NumerationType.Authorization,
            datetime(2019, 1, 1),
            datetime(2019, 5, 1),
        )

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual("authorization", self.qs(last_request)["type"][0])
        self.assertEqual("01/01/2019", self.qs(last_request)["date_start"][0])
        self.assertEqual("01/05/2019", self.qs(last_request)["date_end"][0])

        self.assertEqual(1, len(numerations))
        self.assertEqual("Antioquia", numerations[0].department)

    @respx.mock
    async def test_get_numeration_no_pdf(self):
        self.mock_get_request(respx, "/numeration/", "numeration_no_pdf")
        numerations = await self.client.dian.get_numeration(
            self.session_key,
            NumerationType.Authorization,
            datetime(2019, 1, 1),
            datetime(2019, 5, 1),
        )

        self.assertEqual(1, len(numerations))
        self.assertEqual(None, numerations[0].department)

    @respx.mock
    async def test_retentions(self):
        self.mock_get_request(respx, "/retentions/", "retentions")
        retentions = await self.client.dian.get_retentions(
            self.session_key, "2018", MonthlyPeriod.NOVEMBER
        )

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual("2018", self.qs(last_request)["year"][0])
        self.assertEqual("11", self.qs(last_request)["period"][0])

        self.assertEqual("Qualia Fintech SRL", retentions.reason)

    @respx.mock
    async def test_download_retentions(self):
        self.mock_pdf_download(respx, "/retentions/", "retentions")
        retentions = await self.client.dian.get_retentions(
            self.session_key, "2018", MonthlyPeriod.NOVEMBER
        )
        await self.assert_pdf_downloaded(retentions)



from typing import Optional
from pydantic import BaseModel


class AccountData(BaseModel):
    valid: bool
    message: Optional[str]
    account_number: Optional[str]
    bank_code: Optional[str]
    country_code: Optional[str]
    branch_code: Optional[str]
    document_type: Optional[str]
    document_number: Optional[str]
    beneficiary_name: Optional[str]
    account_currency: Optional[str]
    account_type: Optional[str]



import respx

from prometeo import exceptions
from prometeo.account_validation import exceptions as av_exceptions
from tests.base_test_case import BaseTestCase


class TestAccountValidationClient(BaseTestCase):
    @respx.mock
    def test_invalid_parameters(self):
        self.mock_post_request(respx, "/validate-account/", "invalid_parameters")

        with self.assertRaises(exceptions.InvalidParameterError) as e:
            self.client.account_validation.validate(
                account_number="9999",
                country_code="BR",
                document_number="58.547.642/0001-95",
                branch_code="00001",
                bank_code="999",
                account_type="CHECKING",
            )

        self.assertIn("account_number", e.exception.params)

    @respx.mock
    def test_invalid_parameters_many_errors(self):
        self.mock_post_request(respx, "/validate-account/", "invalid_parameters_many")

        with self.assertRaises(exceptions.InvalidParameterError) as e:
            self.client.account_validation.validate(
                account_number="9999",
                country_code="BR",
                document_number="58.547.642/0001-95",
                branch_code="00001",
                bank_code="999",
                account_type="CHECKING",
            )

        self.assertIn("account_number", e.exception.params)
        self.assertIn("Some description indicating error", e.exception.message)

    @respx.mock
    def test_invalid_account(self):
        self.mock_post_request(respx, "/validate-account/", "invalid_account")

        with self.assertRaises(av_exceptions.InvalidAccountError):
            self.client.account_validation.validate(
                account_number="9999",
                country_code="BR",
                document_number="58.547.642/0001-95",
                branch_code="00001",
                bank_code="999",
                account_type="CHECKING",
            )

    @respx.mock
    def test_validate_accounts(self):
        self.mock_post_request(respx, "/validate-account/", "valid_account_br")
        data = self.client.account_validation.validate(
            account_number="***",
            country_code="BR",
        )
        self.assertEqual(data.beneficiary_name, "JOÃO DAS NEVES")

        self.mock_post_request(respx, "/validate-account/", "valid_account_mx")
        data = self.client.account_validation.validate(
            account_number="***",
            country_code="MX",
        )
        self.assertEqual(data.beneficiary_name, "CRUZ ROJA MEXICANA")

        self.mock_post_request(respx, "/validate-account/", "valid_account_pe")
        data = self.client.account_validation.validate(
            account_number="***",
            country_code="PE",
        )
        self.assertEqual(data.beneficiary_name, "VILLANCA ROSALES ANDREA CLAUDIA")

        self.mock_post_request(respx, "/validate-account/", "valid_account_uy")
        data = self.client.account_validation.validate(
            account_number="***",
            country_code="UY",
        )
        self.assertEqual(data.beneficiary_name, "DANI*** DEC*** COL**")

    @respx.mock
    def test_communication_error(self):
        with self.assertRaises(av_exceptions.CommunicationError):
            self.mock_post_request(respx, "/validate-account/", "communication_error")
            self.client.account_validation.validate(
                account_number="9999",
                country_code="BR",
                document_number="58.547.642/0001-95",
                branch_code="00001",
                bank_code="999",
                account_type="CHECKING",
            )

    @respx.mock
    def test_method_not_available_error(self):
        with self.assertRaises(av_exceptions.MethodNotAvailableError):
            self.mock_post_request(
                respx, "/validate-account/", "method_not_available_error"
            )
            self.client.account_validation.validate(
                account_number="9999",
                country_code="BR",
                document_number="58.547.642/0001-95",
                branch_code="00001",
                bank_code="999",
                account_type="CHECKING",
            )

    @respx.mock
    def test_bank_provider_not_available_error(self):
        with self.assertRaises(av_exceptions.BankProviderNotAvailableError):
            self.mock_post_request(
                respx, "/validate-account/", "bank_not_available_error"
            )
            self.client.account_validation.validate(
                account_number="9999",
                country_code="BR",
                document_number="58.547.642/0001-95",
                branch_code="00001",
                bank_code="999",
                account_type="CHECKING",
            )

    @respx.mock
    def test_country_not_available_error(self):
        with self.assertRaises(av_exceptions.CountryNotAvailableError):
            self.mock_post_request(
                respx, "/validate-account/", "country_not_available_error"
            )
            self.client.account_validation.validate(
                account_number="9999",
                country_code="BR",
                document_number="58.547.642/0001-95",
                branch_code="00001",
                bank_code="999",
                account_type="CHECKING",
            )



from six.moves.urllib.parse import urljoin
import httpx

from prometeo import exceptions, utils


class BaseClient(object):
    """
    Base client class to make api calls
    """

    ENVIRONMENTS = {}

    session_class = None

    def __init__(self, api_key, environment, raw_responses=False, proxy=None):
        self._api_key = api_key
        if environment not in self.ENVIRONMENTS:
            valid_envs = ", ".join(self.ENVIRONMENTS.keys())
            raise exceptions.ClientError(
                'Invalid environment "{}", options are {}'.format(
                    environment, valid_envs
                )
            )
        self._environment = environment
        self._client_session = httpx.AsyncClient(proxy=proxy)
        self._raw_responses = raw_responses

    @utils.adapt_async_sync
    async def make_request(self, method, url, headers=None, *args, **kwargs):
        base_url = self.ENVIRONMENTS[self._environment]
        full_url = urljoin(base_url, url)
        headers = headers or {}
        headers["X-API-Key"] = self._api_key
        return await self._client_session.request(
            method, full_url, headers=headers, *args, **kwargs
        )

    def on_response(self, response_data):
        """
        Called after every 200 response
        """
        pass

    def on_error(self, response, data):
        """
        Check errors on response
        """
        if response.status_code == 400:
            raise exceptions.BadRequestError(data.get("message"))
        elif response.status_code == 401:
            raise exceptions.UnauthorizedError(data.get("message"))
        elif response.status_code == 403 and data.get("status") != "wrong_credentials":
            raise exceptions.ForbiddenError(data.get("message"))
        elif response.status_code == 404:
            raise exceptions.NotFoundError(data.get("message"))
        elif response.status_code == 500:
            raise exceptions.InternalAPIError(data.get("message", response.text))
        elif response.status_code == 503:
            raise exceptions.ProviderUnavailableError(data.get("message"))

    @utils.adapt_async_sync
    async def call_api(self, method, url, headers=None, *args, **kwargs):
        """
        Calls an API endpoint, using the configured api key and environment.

        :param method: The HTTP method to use (``GET``, ``POST``, etc)
        :type method: str

        :param url: The url to call (without the environment's domain)
        :type url: str

        :rtype: JSON data as a python object.
        """
        response = await self.make_request(method, url, headers, *args, **kwargs)
        if not self._raw_responses:
            try:
                data = response.json()
            except ValueError:
                data = {}
            self.on_error(response, data)
            self.on_response(data)
            return data
        return response

    def get_session(self, session_key=""):
        """
        Restore a session from its session key
        :param session_key: The session key
        :type session_key: str

        :rtype: :class:`Session`
        """
        return self.session_class(self, None, session_key)

    def new_session(self):
        """
        Creates a new session

        :param session_key: The session key
        :type session_key: str

        :rtype: :class:`Session`
        """
        return self.session_class(self, None, "")


class Download(object):
    """
    Represents a downloadable file, like an xml bill or pdf document
    """

    def __init__(self, client, url):
        self._client = client
        self.url = url

    @utils.adapt_async_sync
    async def get_file(self):
        """
        Downloads the file and returns its contents.

        :rtype: bytes
        """
        resp = await self._client.make_request("GET", self.url)
        return resp.content



from prometeo.banking.client import BankingAPIClient, Session
from tests.base_test_case import BaseTestCase
import respx


class TestSession(BaseTestCase):
    def setUp(self):
        super(TestSession, self).setUp()
        client = BankingAPIClient("test_api_key", "sandbox")
        self.session_key = "test_session_key"
        self.session = Session(client, "logged_in", self.session_key)

    @respx.mock
    def test_get_accounts(self):
        self.mock_get_request(respx, "/account/", "get_accounts")
        accounts = self.session.get_accounts()
        last_request = respx.calls.last.request
        self.assertEqual(2, len(accounts))
        self.assertEqual(self.session_key, last_request.headers["X-Session-Key"])

    @respx.mock
    def test_get_credit_cards(self):
        self.mock_get_request(respx, "/credit-card/", "get_credit_cards")
        cards = self.session.get_credit_cards()
        last_request = respx.calls.last.request
        self.assertEqual(1, len(cards))
        self.assertEqual(self.session_key, last_request.headers["X-Session-Key"])

    @respx.mock
    def test_restore_session(self):
        self.mock_get_request(respx, "/account/", "get_accounts")

        session_key = "test_restored_key"
        session = self.client.banking.get_session(session_key)

        session.get_accounts()
        last_request = respx.calls.last.request
        self.assertEqual(session_key, last_request.headers["X-Session-Key"])

    @respx.mock
    def test_logout(self):
        self.mock_get_request(respx, "/logout/", "successful_logout")

        self.session.logout()
        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, last_request.headers["X-Session-Key"])

    @respx.mock
    def test_preprocess_transfer(self):
        self.mock_post_request(respx, "/transfer/preprocess", "preprocess_transfer")
        origin_account = "002206345988"
        destination_institution = "0"
        destination_account = "001002363321"
        currency = "UYU"
        amount = "1.3"
        concept = "descripcion de transferencia"
        destination_owner_name = "John Doe"
        branch = "62"
        self.session.preprocess_transfer(
            origin_account,
            destination_institution,
            destination_account,
            currency,
            amount,
            concept,
            destination_owner_name,
            branch,
        )
        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, last_request.headers["X-Session-Key"])

    @respx.mock
    def test_confirm_transfer(self):
        self.mock_post_request(respx, "/transfer/confirm", "confirm_transfer")
        request_id = "0b7d6b32d1be4c11bde21e7ddc08cc36"
        authorization_type = "cardCode"
        authorization_data = "1, 2, 3"
        self.session.confirm_transfer(
            request_id, authorization_type, authorization_data
        )
        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, last_request.headers["X-Session-Key"])

    @respx.mock
    def test_list_transfer_institutions(self):
        self.mock_get_request(
            respx, "/transfer/destinations", "list_transfer_institutions"
        )
        self.session.list_transfer_institutions()
        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, last_request.headers["X-Session-Key"])



class PrometeoError(Exception):
    def __init__(self, message):
        self.message = message


class ClientError(PrometeoError):
    pass


class WrongCredentialsError(PrometeoError):
    pass


class UnauthorizedError(PrometeoError):
    pass


class ForbiddenError(PrometeoError):
    pass


class BadRequestError(PrometeoError):
    pass


class NotFoundError(PrometeoError):
    pass


class InternalAPIError(PrometeoError):
    pass


class ProviderUnavailableError(PrometeoError):
    pass


class InvalidSessionKeyError(PrometeoError):
    pass


class InvalidParameterError(PrometeoError):
    def __init__(self, params, message):
        self.params = params
        self.message = message



from .client import (
    DianAPIClient,
    DocumentType,
    Periodicity,
    QuarterlyPeriod,
    BimonthlyPeriod,
    NumerationType,
    MonthlyPeriod,
)

__all__ = [
    "DianAPIClient",
    "DocumentType",
    "Periodicity",
    "QuarterlyPeriod",
    "BimonthlyPeriod",
    "MonthlyPeriod",
    "NumerationType",
]



from .client import (
    SatAPIClient,
    LoginScope,
    BillStatus,
    Motive,
    DocumentType,
    Status,
    SendType,
)

__all__ = [
    "SatAPIClient",
    "LoginScope",
    "BillStatus",
    "Motive",
    "DocumentType",
    "Status",
    "SendType",
]



from .base_test_case import BaseTestCase
import respx

from prometeo import base_client, exceptions


class TestCase(BaseTestCase):
    def setUp(self):
        self.sandbox_url = "https://test.example.com"
        self.prod_url = "https://prod.example.com"
        base_client.BaseClient.ENVIRONMENTS = {
            "sandbox": self.sandbox_url,
            "production": self.prod_url,
        }
        self.api_key = "test_api_key"
        self.environment = "sandbox"
        self.client = base_client.BaseClient(self.api_key, self.environment)

    @respx.mock
    async def test_api_key(self):
        self.mock_get_request(respx, "/test/", json={})
        await self.client.call_api("GET", "/test/")
        request = respx.calls[0].request
        self.assertIn("X-API-Key", request.headers)
        self.assertEqual(self.api_key, request.headers["X-API-Key"])

    @respx.mock
    async def test_envs(self):
        self.mock_get_request(respx, "/test/", json={})
        history = respx.calls

        test_client = base_client.BaseClient(self.api_key, "sandbox")
        await test_client.call_api("GET", "/test/")
        self.assertIn(self.sandbox_url, str(history[-1].request.url))

        prod_client = base_client.BaseClient(self.api_key, "production")
        await prod_client.call_api("GET", "/test/")
        self.assertIn(self.prod_url, str(history[-1].request.url))

    @respx.mock
    def test_invalid_env(self):
        with self.assertRaises(exceptions.ClientError):
            base_client.BaseClient(self.api_key, "invalid")

    @respx.mock
    async def test_unauthorized(self):
        self.mock_get_request(
            respx,
            "/test/",
            status_code=401,
            json={
                "status": "error",
                "message": "Missing API key",
            },
        )
        with self.assertRaises(exceptions.UnauthorizedError):
            await self.client.call_api("GET", "/test/")

    @respx.mock
    async def test_forbidden(self):
        self.mock_get_request(
            respx,
            "/test/",
            status_code=403,
            json={
                "status": "error",
                "message": "Key not found",
            },
        )
        with self.assertRaises(exceptions.ForbiddenError):
            await self.client.call_api("GET", "/test/")

    @respx.mock
    async def test_missing_credentials(self):
        self.mock_post_request(
            respx,
            "/login/",
            status_code=400,
            json={
                "status": "missing_credentials",
                "message": "missing credentials",
            },
        )
        with self.assertRaises(exceptions.BadRequestError):
            await self.client.call_api("POST", "/login/")

    @respx.mock
    async def test_wrong_parameters(self):
        self.mock_get_request(
            respx,
            "/test/",
            status_code=400,
            json={
                "status": "missing_params",
                "message": "missing parameters",
            },
        )
        with self.assertRaises(exceptions.BadRequestError):
            await self.client.call_api("GET", "/test/")

    @respx.mock
    async def test_404_without_json(self):
        self.mock_get_request(respx, "/invalid/", status_code=404, text="Not found.")
        with self.assertRaises(exceptions.NotFoundError):
            await self.client.call_api("GET", "/invalid/")

    @respx.mock
    async def test_internal_error(self):
        error_message = "Server got itself in trouble."
        self.mock_get_request(respx, "/test/", status_code=500, text=error_message)
        with self.assertRaises(exceptions.InternalAPIError) as cm:
            await self.client.call_api("GET", "/test/")

        self.assertEqual(error_message, cm.exception.message)

    @respx.mock
    async def test_provider_unavailable(self):
        self.mock_get_request(
            respx,
            "/test/",
            status_code=503,
            json={
                "status": "provider_unavailable",
                "message": "Provider returned 500 response.",
            },
        )
        with self.assertRaises(exceptions.ProviderUnavailableError):
            await self.client.call_api("GET", "/test/")

    @respx.mock
    async def test_raw_response(self):
        self.mock_get_request(
            respx,
            "/test/",
            json={
                "name": "test",
                "aliases": ["alias_test"],
                "country": "UY",
                "endpoints_status": None,
                "account_type": [
                    {
                        "name": "pers",
                        "label_es": "Cuenta Personal",
                        "label_en": "Personal Account",
                    },
                    {
                        "name": "corp",
                        "label_es": "Cuenta Corporativa",
                        "label_en": "Corporate Account",
                    },
                ],
            },
        )
        client = base_client.BaseClient(self.api_key, self.environment, True)
        respose = await client.call_api("GET", "/test/")
        self.assertEqual(200, respose.status_code)
        self.assertEqual("test", respose.json()["name"])
        self.assertEqual("UY", respose.json()["country"])



from datetime import datetime
from enum import Enum

from prometeo import exceptions, base_client, base_session, utils
from .models import (
    CFDIBill,
    CFDIDownloadItem,
    DownloadRequest as DownloadRequestModel,
    PdfFile,
    DownloadFile,
    AcknowledgementResult as AcknowledgementResultModel,
)


PRODUCTION_URL = "https://fiscal.prometeoapi.net"
SANDBOX_URL = "https://fiscal.sandbox.prometeoapi.com"


class LoginScope(Enum):
    CFDI = "cfdi"
    SIAT = "siat"


class DownloadAction(Enum):
    BULK_DOWNLOAD = "bulk_download"
    METADATA_DOWNLOAD = "metadata_download"
    PDF_EXPORT = "pdf_export"
    LIST = "list"


class BillStatus(Enum):
    ANY = "any"
    CANCELLED = "cancelled"
    VALID = "valid"


class Motive(Enum):
    ALL = "all"
    AF = "af"
    DE = "de"
    CO = "co"
    FC = "fc"
    MONTHLY = "monthly"


class DocumentType(Enum):
    ALL = "all"
    CT = "ct"
    B = "b"
    PL = "pl"
    XF = "xf"
    XC = "xc"


class Status(Enum):
    ALL = "all"
    RECEIVED = "received"
    ACCEPTED = "accepted"
    REJECTED = "rejected"


class SendType(Enum):
    ALL = "all"
    N = "n"
    C = "c"


class Session(base_session.BaseSession):
    @utils.adapt_async_sync
    async def logout(self):
        """
        Logs out of SAT. You won't be able to use this session after logout.
        """
        await self._client.logout(self._session_key)

    @utils.adapt_async_sync
    async def get_emitted_bills(self, date_start, date_end, status):
        """
        List all emitted bills in a range of dates.

        :param date_start: Start date to filter
        :type date_start: :class:`~datetime.datetime`

        :param date_end: End date to filter
        :type date_end: :class:`~datetime.datetime`

        :param status: Status of the bills
        :type status: :class:`BillStatus`

        :rtype: List of :class:`~prometeo.sat.models.CFDIBill`
        """
        return await self._client.get_emitted(
            self._session_key, date_start, date_end, status, DownloadAction.LIST
        )

    @utils.adapt_async_sync
    async def download_emitted_bills(self, date_start, date_end, status):
        """
        Creates a request to download all the emitted bills in a range of dates.

        :param date_start: Start date to filter
        :type date_start: :class:`~datetime.datetime`

        :param date_end: End date to filter
        :type date_end: :class:`~datetime.datetime`

        :param status: Status of the bills
        :type status: :class:`BillStatus`

        :rtype: List of :class:`~DownloadRequest`
        """
        requests = await self._client.get_emitted(
            self._session_key,
            date_start,
            date_end,
            status,
            DownloadAction.BULK_DOWNLOAD,
        )
        return [
            DownloadRequest(
                self._client,
                self._session_key,
                request.request_id,
            )
            for request in requests
        ]

    @utils.adapt_async_sync
    async def get_received_bills(self, year, month, status):
        """
        List all received bills in a range of dates.

        :param year: Year of the received bills
        :type year: int

        :param month: Month of the received bills
        :type month: int

        :param status: Status of the bills
        :type status: :class:`BillStatus`

        :rtype: List of :class:`~prometeo.sat.models.CFDIBill`
        """
        return await self._client.get_received(
            self._session_key, year, month, status, DownloadAction.LIST
        )

    @utils.adapt_async_sync
    async def download_received_bills(self, year, month, status):
        """
        Creates a request to download all the received bills in a range of dates.

        :param date_start: Start date to filter
        :type date_start: :class:`~datetime.datetime`

        :param date_end: End date to filter
        :type date_end: :class:`~datetime.datetime`

        :param status: Status of the bills
        :type status: :class:`BillStatus`

        :rtype: List of :class:`~DownloadRequest`
        """
        requests = await self._client.get_received(
            self._session_key, year, month, status, DownloadAction.BULK_DOWNLOAD
        )
        return [
            DownloadRequest(
                self._client,
                self._session_key,
                request.request_id,
            )
            for request in requests
        ]

    @utils.adapt_async_sync
    async def get_downloads(self):
        """
        Gets a list of available downloads

        :rtype: List of :class:`DownloadRequest`
        """
        requests = await self._client.get_downloads(self._session_key)
        return [
            DownloadRequest(
                self._client,
                self._session_key,
                request.request_id,
            )
            for request in requests
        ]

    @utils.adapt_async_sync
    async def get_acknowledgements(
        self, year, month_start, month_end, motive, document_type, status, send_type
    ):
        """
        Gets a list of acknowledgements for a range of dates

        :param year: The year of the acknowledgements
        :type year: int

        :param month_start: Start month to filter
        :type month_start: int

        :param month_end: End month to filter
        :type month_end: int

        :param motive: Motive
        :type motive: :class:`Motive`

        :param document_type: Document type
        :type document_type: :class:`DocumentType`

        :param status: Status
        :type status: :class:`Status`

        :param send_type: Send type
        :type send_type: :class:`SendType`

        :rtype: List of :class:`AcknowledgementResult`
        """
        acks = await self._client.get_acknowledgements(
            self._session_key,
            year,
            month_start,
            month_end,
            motive,
            document_type,
            status,
            send_type,
        )
        return [
            AcknowledgementResult(self._client, self._session_key, ack) for ack in acks
        ]


class SatAPIClient(base_client.BaseClient):
    """
    API Client for SAT API
    """

    ENVIRONMENTS = {
        "production": PRODUCTION_URL,
        "sandbox": SANDBOX_URL,
    }

    session_class = Session

    @utils.adapt_async_sync
    async def login(self, rfc, password, scope):
        """
        Log in to SAT

        :param rfc: RFC of the person to log in
        :type rfc: str

        :param password: Password used to login, also known as CIEC
        :type password: str

        :param scope: Depending on the type of information to query, use
                      ``LoginScope.CFDI`` to download bill xmls or
                      ``LoginScope.SIAT`` to download acknowledgements.
        :type scope: :class:`LoginScope`

        :rtype: :class:`Session`
        """
        response = await self.call_api(
            "POST",
            "/login/",
            data={
                "provider": "sat",
                "rfc": rfc,
                "password": password,
                "scope": scope.value,
            },
        )
        if response["status"] == "logged_in":
            return Session(self, response["status"], response["session_key"])
        elif response["status"] == "wrong_credentials":
            raise exceptions.WrongCredentialsError(response["message"])
        else:
            raise exceptions.ClientError(response["message"])

    @utils.adapt_async_sync
    async def logout(self, session_key):
        await self.call_api(
            "GET",
            "/logout/",
            params={
                "session_key": session_key,
            },
        )

    def _handle_bill_parsing(self, action, data):
        if action == DownloadAction.LIST:
            return [
                CFDIBill(
                    id=bill["id"],
                    emitter_rfc=bill["emitter_rfc"],
                    emitter_reason=bill["emitter_reason"],
                    receiver_rfc=bill["receiver_rfc"],
                    receiver_reason=bill["receiver_reason"],
                    emitted_date=datetime.strptime(
                        bill["emitted_date"], "%Y-%m-%dT%H:%M:%S"
                    ),
                    certification_date=datetime.strptime(
                        bill["certification_date"], "%Y-%m-%dT%H:%M:%S"
                    ),
                    certification_pac=bill["certification_pac"],
                    total_value=bill["total_value"],
                    effect=bill["effect"],
                    status=BillStatus(bill["status"]),
                )
                for bill in data
            ]
        elif action == DownloadAction.BULK_DOWNLOAD:
            return [
                DownloadRequestModel(request_id=request["request_id"])
                for request in data
            ]
        elif action == DownloadAction.METADATA_DOWNLOAD:
            return [
                DownloadRequestModel(request_id=request["request_id"])
                for request in data
            ]
        elif action == DownloadAction.PDF_EXPORT:
            return [PdfFile(pdf_url=download["pdf_url"]) for download in data]

    @utils.adapt_async_sync
    async def get_emitted(self, session_key, date_start, date_end, status, action):
        data = await self.call_api(
            "GET",
            "/cfdi/emitted/",
            params={
                "session_key": session_key,
                "date_start": date_start.strftime("%d/%m/%Y"),
                "date_end": date_end.strftime("%d/%m/%Y"),
                "status": status.value,
                "action": action.value,
            },
        )
        return self._handle_bill_parsing(action, data["emitted"])

    @utils.adapt_async_sync
    async def download_emitted(self, session_key, bill_id):
        data = await self.call_api(
            "GET",
            "/cfdi/emitted/{}/".format(bill_id),
            params={
                "session_key": session_key,
            },
        )
        return DownloadFile(**data["download"])

    @utils.adapt_async_sync
    async def get_received(self, session_key, year, month, status, action):
        data = await self.call_api(
            "GET",
            "/cfdi/received/",
            params={
                "session_key": session_key,
                "year": year,
                "month": month,
                "status": status.value,
                "action": action.value,
            },
        )
        return self._handle_bill_parsing(action, data["received"])

    @utils.adapt_async_sync
    async def download_received(self, session_key, bill_id):
        data = await self.call_api(
            "GET",
            "/cfdi/received/{}/".format(bill_id),
            params={
                "session_key": session_key,
            },
        )
        return DownloadFile(**data["download"])

    @utils.adapt_async_sync
    async def get_downloads(self, session_key):
        data = await self.call_api(
            "GET",
            "/cfdi/download/",
            params={
                "session_key": session_key,
            },
        )
        return [CFDIDownloadItem(**item) for item in data["downloads"]]

    @utils.adapt_async_sync
    async def get_download(self, session_key, request_id):
        data = await self.call_api(
            "GET",
            "/cfdi/download/{}/".format(request_id),
            params={
                "session_key": session_key,
            },
        )
        return DownloadFile(**data["download"])

    @utils.adapt_async_sync
    async def get_acknowledgements(
        self,
        session_key,
        year,
        month_start,
        month_end,
        motive,
        document_type,
        status,
        send_type,
    ):
        data = await self.call_api(
            "GET",
            "/ccee/acknowledgment/",
            params={
                "session_key": session_key,
                "year": year,
                "month_start": month_start,
                "month_end": month_end,
                "motive": motive.value,
                "document_type": document_type.value,
                "status": status.value,
                "send_type": send_type.value,
            },
        )
        return [AcknowledgementResultModel(**result) for result in data["results"]]

    @utils.adapt_async_sync
    async def download_acknowledgement(self, session_key, ack_id):
        data = await self.call_api(
            "GET",
            "/ccee/acknowledgment/{}/".format(ack_id),
            params={
                "session_key": session_key,
            },
        )
        return DownloadFile(**data["download"])


class AcknowledgementResult(object):
    """
    Info on an acknowledgement, returned by :meth:`Session.get_acknowledgements`
    """

    def __init__(self, client, session_key, data):
        self._client = client
        self._session_key = session_key
        self.id = data.id
        self.period = data.period
        self.motive = data.motive
        self.document_type = data.document_type
        self.send_type = data.send_type
        self.file_name = data.file_name
        self.reception_date = data.reception_date
        self.status = data.status

    @utils.adapt_async_sync
    async def get_download(self):
        """
        Download the acknowledgement data

        :rtype: :class:`Download`
        """
        download = await self._client.download_acknowledgement(
            self._session_key, self.id
        )
        return base_client.Download(self._client, download.download_url)


class DownloadRequest(object):
    """
    A request for bulk downloading of bills.
    """

    def __init__(self, client, session_key, request_id):
        self._download = None
        self._client = client
        self._session_key = session_key
        self.request_id = request_id

    @utils.adapt_async_sync
    async def get_download(self):
        """
        Download the generated zip file with all the xmls

        :rtype: :class:`Download`
        """
        if self._download is None:
            download = await self._client.get_download(
                self._session_key, self.request_id
            )
            self._download = base_client.Download(self._client, download.download_url)
        return self._download

    @utils.adapt_async_sync
    async def is_ready(self):
        """
        Check if the request is ready to download.

        :rtype: bool
        """
        try:
            await self.get_download()
            return True
        except exceptions.NotFoundError:
            return False



from prometeo import exceptions, base_client, utils
from .exceptions import PaymentInvalidParameterClientError
from typing import Optional, List
from .models import CreatePaymentIntentResponse, PaymentIntent


PRODUCTION_URL = "https://payment.prometeoapi.net"


class PaymentAPIClient(base_client.BaseClient):
    """
    API Client for Payment API
    """

    ENVIRONMENTS = {
        "production": PRODUCTION_URL,
    }

    def on_error(self, response, data):
        message = data.get("message")
        if response.status_code == 400 and message:
            raise exceptions.BadRequestError(data.get("message"))
        first_parameter = list(data.keys())[0]
        if response.status_code == 400 and first_parameter:
            raise PaymentInvalidParameterClientError(
                first_parameter, data.get(first_parameter)
            )

    @utils.adapt_async_sync
    async def create_intent(
        self,
        widget_id: str,
        currency: str,
        amount: str,
        external_id: Optional[str] = None,
        concept: Optional[str] = None,
        bank_codes: Optional[List[str]] = None,
    ) -> CreatePaymentIntentResponse:
        """
        Create payment intent.

        :param widget_id: The widget id
        :type widget_id: str

        :param currency: The payment currency
        :type currency: str

        :param amount: The payment amount
        :type amount: str

        :param external_id: The external id from merchant
        :type external_id: str

        :param concept: The payment concept
        :type concept: str

        :param bank_codes: Select Bank Codes to show in Widget
        :type bank_codes: List[str]

        :return: The payment intent id if it was created, None instead.
        :rtype: CreatePaymentIntentResponse
        """
        data = await self.call_api(
            "POST",
            "/api/v1/payment-intent/",
            json={
                "product_id": widget_id,
                "product_type": "widget",
                "currency": currency,
                "amount": amount,
                "external_id": external_id,
                "concept": concept,
                "bank_codes": bank_codes,
            },
        )
        return CreatePaymentIntentResponse(**data)

    @utils.adapt_async_sync
    async def get_transaction_data(self, intent_id: str) -> PaymentIntent:
        """
        Get transaction data from intent.

        :param intent_id: The intent id
        :type intent_id: str

        :return: The payment intent transaction data
        :rtype: PaymentIntent
        """
        data = await self.call_api("GET", f"/api/v1/payment-intent/{intent_id}")
        return PaymentIntent(**data)



import asyncio
import functools


def adapt_async_sync(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if asyncio.iscoroutinefunction(func):
            loop = asyncio.get_event_loop()
            if loop.is_running():
                return func(*args, **kwargs)
            else:
                return loop.run_until_complete(func(*args, **kwargs))
        else:
            return func(*args, **kwargs)

    return wrapper



from .banking import BankingAPIClient
from .curp import CurpAPIClient
from .dian import DianAPIClient
from .sat import SatAPIClient
from .payment import PaymentAPIClient
from .account_validation import AccountValidationAPIClient


class Client(object):
    def __init__(self, api_key, environment="sandbox", raw_responses=False, proxy=None):
        self._api_key = api_key
        self._environment = environment
        self._raw_responses = raw_responses
        self._proxy = proxy
        self._banking = None
        self._dian = None
        self._sat = None
        self._curp = None
        self._payment = None
        self._account_validation = None

    @property
    def banking(self):
        if self._banking is None:
            self._banking = BankingAPIClient(
                self._api_key, self._environment, self._raw_responses, self._proxy
            )
        return self._banking

    @property
    def dian(self):
        if self._dian is None:
            self._dian = DianAPIClient(
                self._api_key, self._environment, self._raw_responses, self._proxy
            )
        return self._dian

    @property
    def sat(self):
        if self._sat is None:
            self._sat = SatAPIClient(
                self._api_key, self._environment, self._raw_responses, self._proxy
            )
        return self._sat

    @property
    def curp(self):
        if self._curp is None:
            self._curp = CurpAPIClient(
                self._api_key, self._environment, self._raw_responses, self._proxy
            )
        return self._curp

    @property
    def payment(self):
        if self._payment is None:
            self._payment = PaymentAPIClient(
                self._api_key, self._environment, self._raw_responses, self._proxy
            )
        return self._payment

    @property
    def account_validation(self):
        if self._account_validation is None:
            self._account_validation = AccountValidationAPIClient(
                self._api_key, self._environment, self._raw_responses, self._proxy
            )
        return self._account_validation



import json
import asynctest

from prometeo import Client
import httpx
from six.moves.urllib.parse import parse_qs, urlparse


class BaseTestCase(asynctest.TestCase):
    def setUp(self):
        self.client = Client("test_key")

    @property
    def module_name(self):
        return self.__class__.__module__.split(".")[1]

    def load_json(self, file_name):
        with open(f"tests/fixtures/{self.module_name}/{file_name}.json") as f:
            return json.load(f)

    def mock_get_request(
        self, mocker, url, fixture_name=None, json=None, status_code=200, **kwargs
    ):
        if not fixture_name:
            mocker.get(url).mock(
                return_value=httpx.Response(status_code, json=json, **kwargs)
            )
        else:
            mocker.get(url).mock(
                return_value=httpx.Response(
                    status_code, json=self.load_json(fixture_name)
                )
            )

    def mock_post_request(
        self, mocker, url, fixture_name=None, json=None, status_code=200, **kwargs
    ):
        if not fixture_name:
            mocker.post(url).mock(
                return_value=httpx.Response(status_code, json=json, **kwargs)
            )
        else:
            mocker.post(url).mock(
                return_value=httpx.Response(
                    status_code, json=self.load_json(fixture_name)
                )
            )

    def qs(self, last_request):
        return parse_qs(urlparse(str(last_request.url)).query)



from prometeo import exceptions


class PaymentClientError(exceptions.PrometeoError):
    pass


class PaymentInvalidParameterClientError(PaymentClientError):
    def __init__(self, param, message):
        self.param = param
        self.message = message



from enum import Enum
from datetime import datetime

from prometeo import base_client, utils
from .models import QueryResult, DocumentData, PersonalData
from .exceptions import CurpError

PRODUCTION_URL = "https://identity.prometeoapi.net"
SANDBOX_URL = "https://identity.sandbox.prometeoapi.com"


class Gender(Enum):
    """A person's gender"""

    MALE = "H"
    FEMALE = "M"


class State(Enum):
    """
    The state a person is registered
    """

    AGUASCALIENTES = "AS"
    BAJA_CALIFORNIA = "BC"
    BAJA_CALIFORNIA_SUR = "BS"
    CAMPECHE = "CC"
    COAHUILA = "CL"
    COLIMA = "CM"
    CHIAPAS = "CS"
    CHIHUAHUA = "CH"
    CIUDAD_DE_MEXICO = "DF"
    DURANGO = "DG"
    GUANAJUATO = "GT"
    GUERRERO = "GR"
    HIDALGO = "HG"
    JALISCO = "JC"
    ESTADO_DE_MEXICO = "MC"
    MICHOACAN = "MN"
    MORELOS = "MS"
    NAYARIT = "NT"
    NUEVO_LEON = "NL"
    OAXACA = "OC"
    PUEBLA = "PL"
    QUERETARO = "QT"
    QUINTANA_ROO = "QR"
    SAN_LUIS_POTOSI = "SP"
    SINALOA = "SL"
    SONORA = "SR"
    TABASCO = "TC"
    TAMAULIPAS = "TS"
    TLAXCALA = "TL"
    VERACRUZ = "VZ"
    YUCATAN = "YN"
    ZACATECA = "ZS"
    NACIDO_EN_EL_EXTRANJERO = "NE"


class CurpAPIClient(base_client.BaseClient):
    """
    API Client for CURP queries
    """

    ENVIRONMENTS = {
        "production": PRODUCTION_URL,
        "sandbox": SANDBOX_URL,
    }

    def _parse_date(self, date_str):
        if not date_str:
            return None
        for fmt in ("%Y-%m-%d", "%d/%m/%Y"):
            try:
                return datetime.strptime(date_str, fmt)
            except ValueError:
                continue
        return None

    def _make_result(self, response_data):
        document_data = response_data.get("document_data", {})
        personal_data = response_data.get("personal_data", {})
        raw_birthdate = personal_data.get("fechaNacimiento")
        return QueryResult(
            document_data=DocumentData(
                foja=document_data.get("foja"),
                clave_entidad_registro=document_data.get("claveEntidadRegistro"),
                num_acta=document_data.get("numActa"),
                tomo=document_data.get("tomo"),
                anio_reg=document_data.get("anioReg"),
                municipio_registro=document_data.get("municipioRegistro"),
                libro=document_data.get("libro"),
                entidad_registro=document_data.get("entidadRegistro"),
                clave_municipio_registro=document_data.get("claveMunicipioRegistro"),
            ),
            personal_data=PersonalData(
                sexo=personal_data.get("sexo"),
                entidad=personal_data.get("entidad"),
                nacionalidad=personal_data.get("nacionalidad"),
                status_curp=personal_data.get("statusCurp"),
                nombres=personal_data.get("nombres"),
                segundo_apellido=personal_data.get("segundoApellido"),
                clave_entidad=personal_data.get("claveEntidad"),
                doc_probatorio=personal_data.get("docProbatorio"),
                fecha_nacimiento=self._parse_date(raw_birthdate),
                primer_apellido=personal_data.get("primerApellido"),
                curp=personal_data.get("curp"),
            ),
            pdf_url=response_data.get("pdf_url", ""),
            pdf=base_client.Download(self, response_data.get("pdf_url", "")),
        )

    @utils.adapt_async_sync
    async def query(self, curp):
        """
        Find the personal data associated with a CURP

        :param curp: The CURP of the person to query
        :type curp: str

        :rtype: :class:`~prometeo.curp.models.QueryResult`
        """
        response = await self.call_api(
            "POST",
            "/query",
            data={
                "curp": curp,
            },
        )
        if response["errors"] is not None:
            raise CurpError(response["errors"]["detail"])
        return self._make_result(response["data"])

    @utils.adapt_async_sync
    async def reverse_query(
        self, state, birthdate, name, first_surname, last_surname, gender
    ):
        """
        Search for a person by their personal information.

        :param state: The state where the person is registered
        :type state: :class:`State`

        :param birthdate: The person's birthdate
        :type birthdate: :class:`~datetime.datetime`

        :param name: The person's name
        :type name: str

        :param first_surname: The person's first surname
        :type first_surname: str

        :param last_surname: The person's last surname
        :type last_surname: str

        :param gender: The person's gender
        :type gender: :class:`Gender`

        :rtype: :class:`~prometeo.curp.models.QueryResult`
        """
        response = await self.call_api(
            "POST",
            "/reverse-query",
            data={
                "state": state.value,
                "birthdate": birthdate.strftime("%d/%m/%Y"),
                "name": name,
                "first_surname": first_surname,
                "last_surname": last_surname,
                "gender": gender.value,
            },
        )
        if response["errors"] is not None:
            raise CurpError(response["errors"]["detail"])
        return self._make_result(response["data"])



from datetime import datetime

from prometeo.dian.client import (
    DianAPIClient,
    Session,
    Periodicity,
    QuarterlyPeriod,
    NumerationType,
    MonthlyPeriod,
)
from tests.base_test_case import BaseTestCase
import respx


class TestSession(BaseTestCase):
    def setUp(self):
        super(TestSession, self).setUp()
        client = DianAPIClient("test_api_key", "sandbox")
        self.session_key = "test_session_key"
        self.session = Session(client, "logged_in", self.session_key)

    @respx.mock
    def test_get_company_info(self):
        self.mock_get_request(respx, "/company-info/", "company_info")
        info = self.session.get_company_info()

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual(info.reason, "Qualia Fintech SRL")
        self.assertEqual(info.location.country, "Uruguay")
        self.assertEqual(info.accountant.start_date, datetime(2017, 5, 1))

    @respx.mock
    def test_get_balances(self):
        self.mock_get_request(respx, "/balances/", "company_balances")
        balances = self.session.get_balances()
        last_request = respx.calls.last.request
        self.assertEqual(2, len(balances))
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])

    @respx.mock
    def test_get_rent_declaration(self):
        self.mock_get_request(respx, "/rent/", "company_rent")
        rent = self.session.get_rent_declaration(2019)

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual("2019", self.qs(last_request)["year"][0])
        self.assertEqual(rent.nit, "333222251")
        self.assertEqual(6, len(rent.fields))

    @respx.mock
    def test_get_vat_declaration(self):
        self.mock_get_request(respx, "/vat/", "vat_declaration")
        vat = self.session.get_vat_declaration(
            2019,
            Periodicity.QUARTERLY,
            QuarterlyPeriod.JANUARY_APRIL,
        )

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual("2019", self.qs(last_request)["year"][0])
        self.assertEqual("q", self.qs(last_request)["periodicity"][0])
        self.assertEqual("1", self.qs(last_request)["period"][0])
        self.assertEqual(vat.nit, "123332211")
        self.assertEqual(6, len(vat.fields))

    @respx.mock
    def test_get_numeration(self):
        self.mock_get_request(respx, "/numeration/", "numeration_authorization")
        numerations = self.session.get_numeration(
            NumerationType.Authorization,
            datetime(2019, 1, 1),
            datetime(2019, 5, 1),
        )

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual("authorization", self.qs(last_request)["type"][0])
        self.assertEqual("01/01/2019", self.qs(last_request)["date_start"][0])
        self.assertEqual("01/05/2019", self.qs(last_request)["date_end"][0])

        self.assertEqual(1, len(numerations))
        self.assertEqual("Antioquia", numerations[0].department)

    @respx.mock
    def test_retentions(self):
        self.mock_get_request(respx, "/retentions/", "retentions")
        retentions = self.session.get_retentions(2018, MonthlyPeriod.NOVEMBER)

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual("2018", self.qs(last_request)["year"][0])
        self.assertEqual("11", self.qs(last_request)["period"][0])

        self.assertEqual("Qualia Fintech SRL", retentions.reason)

    @respx.mock
    def test_restore_session(self):
        self.mock_get_request(respx, "/balances/", "company_balances")

        session_key = "test_restored_key"
        session = self.client.dian.get_session(session_key)

        session.get_balances()

        last_request = respx.calls.last.request
        self.assertEqual(session_key, self.qs(last_request)["session_key"][0])



from enum import Enum
from datetime import datetime

from prometeo import exceptions, base_client, base_session, utils
from .models import (
    CompanyInfo,
    Member,
    Representative,
    Name,
    Accountant,
    CapitalComposition,
    Location,
    Balance,
    RentDeclaration,
    Field,
    VATDeclaration,
    Numeration,
    NumerationRange,
    Retentions,
)

PRODUCTION_URL = "https://fiscal.prometeoapi.net"
SANDBOX_URL = "https://fiscal.sandbox.prometeoapi.com"


class NumerationType(Enum):
    Authorization = "authorization"
    Habilitation = "habilitation"
    Inhabilitation = "inhabilitation"


class Periodicity(Enum):
    QUARTERLY = "q"
    BIMONTHLY = "b"


class QuarterlyPeriod(Enum):
    JANUARY_APRIL = 1
    MAY_AUGUST = 2
    SEPTEMBER_DECEMBER = 3


class BimonthlyPeriod(Enum):
    JANUARY_FEBRUARY = 1
    MARCH_APRIL = 2
    MAY_JUNE = 3
    JULY_AUGUST = 4
    SEPTEMBER_OCTOBER = 5
    NOVEMBER_DECEMBER = 6


class MonthlyPeriod(Enum):
    JANUARY = 1
    FEBRUARY = 2
    MARCH = 3
    APRIL = 4
    MAY = 5
    JUNE = 6
    JULY = 7
    AUGUST = 8
    SEPTEMBER = 9
    OCTOBER = 10
    NOVEMBER = 11
    DECEMBER = 12


class DocumentType(Enum):
    """
    Document types used for :meth:`DianAPIClient.login`
    """

    TARJETA_IDENTIDAD = "12"
    CEDULA_CIUDADANIA = "13"
    CERTIFICADO_REGISTRADURIA = "14"
    TARJETA_EXTRANJERIA = "21"
    CEDULA_EXTRANJERIA = "22"
    PASAPORTE = "41"
    DOCUMENTO_IDENTIFICACION_EXTRANJERO = "42"
    SIN_IDENTIFICACION_EXTRANJERO = "43"
    DOCUMENTO_IDENTIFICACION_EXTRANJERO_PERSONA_JURIDICA = "44"
    CARNE_DIPLOMATICO = "46"


class Session(base_session.BaseSession):
    @utils.adapt_async_sync
    async def get_company_info(self):
        """
        Get information about the company or person (form 001).

        :rtype: :class:`~prometeo.dian.models.CompanyInfo`
        """
        return await self._client.get_company_info(self._session_key)

    @utils.adapt_async_sync
    async def get_balances(self):
        """
        Get balances.

        :rtype: list of :class:`~prometeo.dian.models.Balance`
        """
        return await self._client.get_balances(self._session_key)

    @utils.adapt_async_sync
    async def get_rent_declaration(self, year):
        """
        Get rent declaration information (form 110 for companies and 210 for persons)

        :param year: Year of the declaration
        :type year: int

        :rtype: :class:`~prometeo.dian.models.RentDeclaration`
        """
        return await self._client.get_rent_declaration(self._session_key, year)

    @utils.adapt_async_sync
    async def get_vat_declaration(self, year, periodicity, period):
        """
        Get VAT declaration information (form 300)

        :param year: Year of the declaration
        :type year: int

        :param periodicity: Periodicity, either quarterly or bimonthly
        :type periodicity: :class:`Periodicity`

        :param period: Period of the declaration
        :type period: int

        :rtype: :class:`~prometeo.dian.models.VATDeclaration`
        """
        return await self._client.get_vat_declaration(
            self._session_key, year, periodicity, period
        )

    @utils.adapt_async_sync
    async def get_numeration(self, type, date_start, date_end):
        """
        Get bill numeration (form 1876)

        :param type: The type of numeration request to get
        :type type: :class:`NumerationType`

        :param date_start: Start date to filter
        :type date_start: :class:`~datetime.datetime`

        :param date_end: End date to filter
        :type date_end: :class:`~datetime.datetime`

        :rtype: List of :class:`~prometeo.dian.models.Numeration`
        """
        return await self._client.get_numeration(
            self._session_key, type, date_start, date_end
        )

    @utils.adapt_async_sync
    async def get_retentions(self, year, period):
        """
        Get retentions information (form 350)

        :param year: Year of the retention
        :type year: int

        :param period: Period of the retention
        :type period: int

        :rtype: List of :class:`~prometeo.dian.models.Retentions`
        """
        return await self._client.get_retentions(self._session_key, year, period)


class DianAPIClient(base_client.BaseClient):
    """
    API Client for DIAN API
    """

    ENVIRONMENTS = {
        "production": PRODUCTION_URL,
        "sandbox": SANDBOX_URL,
    }

    session_class = Session

    @utils.adapt_async_sync
    async def login(self, document_type, document, password, nit=None):
        """
        Log in to DIAN

        :param document_type: The type of document for the ``document`` argument.
        :type document_type: :class:`DocumentType`

        :param document: The document of the person
        :type document: str

        :param password: The password used to log in
        :type password: str

        :param nit: The NIT of the company (used only when logging in as a company)
        :type nit: str

        :rtype: :class:`Session`
        """
        data = {
            "provider": "dian",
            "document_type": document_type.value,
            "document": document,
            "password": password,
        }
        if nit is not None:
            data["nit"] = nit
        response = await self.call_api("POST", "/login/", data=data)
        if response["status"] == "logged_in":
            return Session(self, response["status"], response["session_key"])
        elif response["status"] == "wrong_credentials":
            raise exceptions.WrongCredentialsError(response["message"])
        else:
            raise exceptions.ClientError(response["message"])

    @utils.adapt_async_sync
    async def get_company_info(self, session_key):
        response = await self.call_api(
            "GET",
            "/company-info/",
            params={
                "session_key": session_key,
            },
        )
        data = response["info"]
        return CompanyInfo(
            accountant=Accountant(
                document=data["accountant"]["document"],
                start_date=datetime.strptime(
                    data["accountant"]["start_date"], "%d/%m/%Y"
                ),
                name=data["accountant"]["name"],
                professional_card=data["accountant"]["professional_card"],
            ),
            capital_composition=CapitalComposition(**data["capital_composition"]),
            reason=data["reason"],
            pdf_url=data["pdf_url"],
            pdf=base_client.Download(self, data["pdf_url"]),
            location=Location(**data["location"]),
            name=data["name"],
            constitution_date=datetime.strptime(data["constitution_date"], "%d/%m/%Y"),
            representation=[
                Representative(
                    representation_type=representative["representation_type"],
                    start_date=datetime.strptime(
                        representative["start_date"], "%d/%m/%Y"
                    ),
                    document_type=representative["document_type"],
                    document=representative["document"],
                    name=Name(**representative["name"]),
                )
                for representative in data["representation"]
            ],
            members=[
                Member(
                    document_type=member["document_type"],
                    document=member["document"],
                    nationality=member["nationality"],
                    name=Name(**member["name"]),
                    start_date=datetime.strptime(member["start_date"], "%d/%m/%Y"),
                )
                for member in data["members"]
            ],
        )

    @utils.adapt_async_sync
    async def get_balances(self, session_key):
        data = await self.call_api(
            "GET",
            "/balances/",
            params={
                "session_key": session_key,
            },
        )
        return [Balance(**balance) for balance in data["balances"]]

    @utils.adapt_async_sync
    async def get_rent_declaration(self, session_key, year):
        data = await self.call_api(
            "GET",
            "/rent/",
            params={
                "session_key": session_key,
                "year": year,
            },
        )
        return RentDeclaration(
            pdf_url=data["declaration"]["pdf_url"],
            pdf=base_client.Download(self, data["declaration"]["pdf_url"]),
            fields=[Field(**field) for field in data["declaration"]["fields"].values()],
            year=data["declaration"]["year"],
            form_number=data["declaration"]["form_number"],
            nit=data["declaration"]["nit"],
            dv=data["declaration"]["dv"],
            name=Name(**data["declaration"]["name"]),
            reason=data["declaration"]["reason"],
            direction_code=data["declaration"]["direction_code"],
            economic_activity=data["declaration"]["economic_activity"],
            correction_code=data["declaration"]["correction_code"],
            previous_form=data["declaration"]["previous_form"],
        )

    @utils.adapt_async_sync
    async def get_vat_declaration(self, session_key, year, periodicity, period):
        data = await self.call_api(
            "GET",
            "/vat/",
            params={
                "session_key": session_key,
                "year": year,
                "periodicity": periodicity.value,
                "period": period.value,
            },
        )
        return VATDeclaration(
            pdf_url=data["declaration"]["pdf_url"],
            pdf=base_client.Download(self, data["declaration"]["pdf_url"]),
            fields=[Field(**field) for field in data["declaration"]["fields"].values()],
            year=data["declaration"]["year"],
            form_number=data["declaration"]["form_number"],
            nit=data["declaration"]["nit"],
            dv=data["declaration"]["dv"],
            name=Name(**data["declaration"]["name"]),
            reason=data["declaration"]["reason"],
            direction_code=data["declaration"]["direction_code"],
            correction_code=data["declaration"]["correction_code"],
            previous_form=data["declaration"]["previous_form"],
            period=data["declaration"]["period"],
        )

    @utils.adapt_async_sync
    async def get_numeration(self, session_key, type, date_start, date_end):
        data = await self.call_api(
            "GET",
            "/numeration/",
            params={
                "session_key": session_key,
                "type": type.value,
                "date_start": date_start.strftime("%d/%m/%Y"),
                "date_end": date_end.strftime("%d/%m/%Y"),
            },
        )
        numerations = []
        for numeration in data["numeration"]:
            if numeration["name"]:
                name = Name(**numeration["name"])
            else:
                name = None
            numerations.append(
                Numeration(
                    address=numeration["address"],
                    country=numeration["country"],
                    department=numeration["department"],
                    dv=numeration["dv"],
                    municipality=numeration["municipality"],
                    nit=numeration["nit"],
                    pdf_url=numeration["pdf_url"],
                    pdf_available=numeration["pdf_available"],
                    reason=numeration["reason"],
                    name=name,
                    ranges=[
                        NumerationRange(
                            establishment=range["establishment"],
                            from_number=range["from"],
                            to_number=range["to"],
                            mode=range["mode"],
                            prefix=range["prefix"],
                            type=range["type"],
                        )
                        for range in numeration["ranges"]
                    ],
                )
            )
        return numerations

    @utils.adapt_async_sync
    async def get_retentions(self, session_key, year, period):
        data = await self.call_api(
            "GET",
            "/retentions/",
            params={
                "session_key": session_key,
                "year": year,
                "period": period.value,
            },
        )
        return Retentions(
            pdf_url=data["retentions"]["pdf_url"],
            pdf=base_client.Download(self, data["retentions"]["pdf_url"]),
            fields=[Field(**field) for field in data["retentions"]["fields"].values()],
            year=data["retentions"]["year"],
            form_number=data["retentions"]["form_number"],
            nit=data["retentions"]["nit"],
            reason=data["retentions"]["reason"],
            direction_code=data["retentions"]["direction_code"],
            period=data["retentions"]["period"],
        )



from typing import Optional, Union
from pydantic import BaseModel
from datetime import datetime

class DocumentData(BaseModel):
    foja: Optional[str] = None
    clave_entidad_registro: str
    num_acta: str
    tomo: Optional[str] = None
    anio_reg: str
    municipio_registro: str
    libro: Optional[str] = None
    entidad_registro: str
    clave_municipio_registro: str

class PersonalData(BaseModel):
    sexo: str
    entidad: str
    nacionalidad: str
    status_curp: str
    nombres: str
    segundo_apellido: str
    clave_entidad: str
    doc_probatorio: Union[int, str]  # Allowing for both int and str types
    fecha_nacimiento: datetime
    primer_apellido: str
    curp: str

class QueryResult(BaseModel):
    document_data: DocumentData
    personal_data: PersonalData
    pdf_url: Optional[str] = None
    pdf: Optional[object] = None



from datetime import datetime

import six
from six.moves.urllib.parse import parse_qs


from prometeo.curp import exceptions, Gender, State
from tests.base_test_case import BaseTestCase
import respx


class TestCurpClient(BaseTestCase):
    @respx.mock
    async def test_query_success(self):
        test_curp = "ABCD880304HDFXNR45"
        self.mock_post_request(respx, "/query", "successful_curp")
        result = await self.client.curp.query(test_curp)

        last_request = respx.calls.last.request
        request_body = parse_qs(str(last_request.content.decode("utf-8")))
        self.assertEqual(test_curp, request_body["curp"][0])
        self.assertEqual("SINALOA", result.document_data.entidad_registro)
        self.assertEqual(datetime(1988, 3, 4), result.personal_data.fecha_nacimiento)

    @respx.mock
    def test_curp_nonexistent(self):
        error_message = (
            "Estimada/o usuaria/o, la clave que "
            "ingresaste no existe o no está validada."
        )
        self.mock_post_request(respx, "/query", "inexistent_curp")

        with self.assertRaises(exceptions.CurpError) as cm:
            self.client.curp.query("ABCD123445")

        self.assertIn(error_message, cm.exception.message)

    @respx.mock
    async def test_reverse_query(self):
        self.mock_post_request(respx, "/reverse-query", "successful_curp")

        state = State.SINALOA
        birthdate = datetime(1988, 3, 4)
        name = "JOHN"
        first_surname = "DOE"
        last_surname = "PONCE"
        gender = Gender.MALE
        result = await self.client.curp.reverse_query(
            state, birthdate, name, first_surname, last_surname, gender
        )

        last_request = respx.calls.last.request
        request_body = parse_qs(str(last_request.content.decode("utf-8")))
        self.assertEqual("SL", request_body["state"][0])
        self.assertEqual("04/03/1988", request_body["birthdate"][0])
        self.assertEqual(name, request_body["name"][0])
        self.assertEqual(first_surname, request_body["first_surname"][0])
        self.assertEqual(last_surname, request_body["last_surname"][0])
        self.assertEqual("H", request_body["gender"][0])

        self.assertEqual("SINALOA", result.document_data.entidad_registro)
        self.assertEqual(datetime(1988, 3, 4), result.personal_data.fecha_nacimiento)

    @respx.mock
    def test_reverse_query_no_result(self):
        error_message = (
            "Estimada/o usuaria/o, la clave que "
            "ingresaste no existe o no está validada."
        )
        self.mock_post_request(respx, "/reverse-query", "inexistent_curp")

        state = State.SINALOA
        birthdate = datetime(1988, 3, 4)
        name = "JOHN"
        first_surname = "DOE"
        last_surname = "PONCE"
        gender = Gender.MALE
        with self.assertRaises(exceptions.CurpError) as cm:
            self.client.curp.reverse_query(
                state, birthdate, name, first_surname, last_surname, gender
            )
        self.assertIn(error_message, cm.exception.message)

    @respx.mock
    async def test_pdf_url(self):
        test_curp = "ABCD880304HDFXNR45"
        self.mock_post_request(respx, "/query", "successful_curp")
        result = await self.client.curp.query(test_curp)
        self.assertEqual("/pdf/50ad2ba127ae4cc384fd265e585a1f67.pdf", result.pdf_url)

    @respx.mock
    def test_download_pdf(self):
        test_curp = "ABCD880304HDFXNR45"
        pdf_url = "/pdf/50ad2ba127ae4cc384fd265e585a1f67.pdf"
        pdf_content = "pdf content"
        self.mock_post_request(respx, "/query", "successful_curp")
        self.mock_get_request(respx, pdf_url, text=pdf_content)

        result = self.client.curp.query(test_curp)
        self.assertEqual(six.ensure_binary(pdf_content), result.pdf.get_file())



from datetime import datetime

from prometeo import exceptions, base_client, base_session, utils
from .models import (
    Client as Client,
    Account as AccountModel,
    Movement,
    CreditCard as CreditCardModel,
    Provider,
    ProviderDetail,
    PreprocessTransfer,
    ConfirmTransfer,
    TransferInstitution,
)
from .exceptions import BankingClientError


PRODUCTION_URL = "https://banking.prometeoapi.net"
SANDBOX_URL = "https://banking.sandbox.prometeoapi.com"


class Session(base_session.BaseSession):
    """
    Encapsulates the user's session, returned by
    :meth:`~prometeo.banking.client.BankingAPIClient.login`
    """

    def __init__(self, client, status, session_key, context=None, field=None):
        super(Session, self).__init__(client, status, session_key)
        self._interactive_context = context
        self._interactive_field = field

    @utils.adapt_async_sync
    async def login(self, provider, username, password, **kwargs):
        """
        Start log in process with the provider

        :param provider: Name of the provider,
                         use :meth:`~BankingAPIClient.get_providers`
                         to get a list of available providers
        :type provider: str

        :param username: Username used to log in to the banking app or web
        :type username: str

        :param password: User's password
        :type password: str

        :param kwargs: Extra login fields for providers that require it,
                       use :meth:`~BankingAPIClient.get_provider_detail`
                       to get a list of all the auth login fields
        :type kwargs: dict

        :rtype: :class:`~prometeo.banking.client.Session`
        """
        data = {
            "provider": provider,
            "username": username,
            "password": password,
            **kwargs,
        }
        response = await self._client.login(**data)
        if response["status"] in ["logged_in", "select_client"]:
            self._session_key = response.get("key") or self._session_key
            self._status = response["status"]
        elif response["status"] == "interaction_required":
            self._session_key = response.get("key")
            self._status = response["status"]
            self._interactive_context = response["context"]
            self._interactive_field = response["field"]
        elif response["status"] == "wrong_credentials":
            raise exceptions.WrongCredentialsError(response["message"])
        else:
            raise BankingClientError(response["message"])

    @utils.adapt_async_sync
    async def finish_login(self, provider, username, password, answer):
        return await self.login(
            provider,
            username,
            password,
            session_key=self._session_key,
            **{self._interactive_field: answer},
        )

    @utils.adapt_async_sync
    async def get_clients(self):
        """
        List the user's clients. Returns an empty list if the bank doesn't uses clients.

        :rtype: List of :class:`~prometeo.banking.models.Client`
        """
        response = await self._client.get_clients(self._session_key)
        clients = []
        for id, name in response["clients"].items():
            clients.append(Client(id=id, name=name))
        return clients

    @utils.adapt_async_sync
    async def select_client(self, client):
        """
        Selects the client to use for this session.

        :param client_id: The id of the client, obtained from listing the clients
        :type client_id: str
        """
        await self._client.select_client(self._session_key, client.id)

    @utils.adapt_async_sync
    async def get_accounts(self):
        """
        List all the user's accounts

        :rtype: List of :class:`~prometeo.banking.models.Account`
        """
        data = await self._client.get_accounts(self._session_key)
        accounts_data = [AccountModel(**account) for account in data["accounts"]]
        accounts = []
        for account_data in accounts_data:
            accounts.append(
                Account(
                    self._client,
                    self._session_key,
                    account_data,
                )
            )
        return accounts

    @utils.adapt_async_sync
    async def get_credit_cards(self):
        """
        List all the user's credit cards

        :rtype: List of :class:`~prometeo.banking.models.CreditCard`
        """
        data = await self._client.get_credit_cards(self._session_key)
        cards_data = [
            CreditCardModel(
                id=credit_card["id"],
                name=credit_card["name"],
                number=credit_card["number"],
                close_date=datetime.strptime(credit_card["close_date"], "%d/%m/%Y"),
                due_date=datetime.strptime(credit_card["due_date"], "%d/%m/%Y"),
                balance_local=credit_card["balance_local"],
                balance_dollar=credit_card["balance_dollar"],
            )
            for credit_card in data["credit_cards"]
        ]
        cards = []
        for card_data in cards_data:
            cards.append(
                CreditCard(
                    self._client,
                    self._session_key,
                    card_data,
                )
            )
        return cards

    @utils.adapt_async_sync
    def get_interactive_context(self):
        """
        Necessary information to answer the login challenge, like a security question.

        :rtype: str
        """
        return self._interactive_context

    @utils.adapt_async_sync
    async def get_providers(self):
        """
        List all available banks.

        :rtype: :class:`~prometeo.banking.models.Provider`
        """
        data = await self._client.get_providers()
        return [Provider(**provider) for provider in data["providers"]]

    @utils.adapt_async_sync
    async def get_provider_detail(self, provider_code, key=None, value=None):
        """
        Get more detailed information about a bank.

        :param provider_code: Name of the provider, as returned in
                              :meth:`~prometeo.banking.client.BankingAPIClient.get_providers`
        :type provider_code: str

        :param key: Auth Field Key (for filtering options)
        :type key: str

        :param value: Auth Field Value (for filtering options)
        :type value: str

        :rtype: :class:`~prometeo.banking.models.ProviderDetail`
        """
        data = await self._client.get_provider_detail(provider_code, key, value)
        return ProviderDetail(**data["provider"])

    @utils.adapt_async_sync
    async def logout(self):
        """
        Logs the user out and invalidates its session.
        """
        await self._client.logout(self._session_key)

    @utils.adapt_async_sync
    async def preprocess_transfer(
        self,
        origin_account,
        destination_institution,
        destination_account,
        currency,
        amount,
        concept,
        destination_owner_name,
        branch,
        destination_account_type=None,
        document_type=None,
        document_number=None,
        country=None,
        destination_bank_code=None,
        payment_intent_id=None,
        external_id=None,
        mobile_os=None,
    ):
        """
        Preprocess transfer.

        :param origin_account: Account number from where it is transferred
        :type origin_account: str

        :param destination_institution: Id of the institution where it is transferred.
                                        As provided by
                                        :meth:`~prometeo.banking.client.BankingAPIClient.get_providers`
        :type destination_institution: str

        :param destination_account: Account number where it is transferred
        :type destination_account: str

        :param currency: Account currency in format ``ISO 4217``
        :type currency: str

        :param amount: Amount to transfer
        :type amount: str

        :param concept: Concept or description of the transfer
        :type concept: str

        :param destination_owner_name:
            Name of the holder of the destination account (empty if not applicable)
        :type destination_owner_name: str

        :param branch: Branch number of the destination account
                       (empty if not applicable)
        :type branch: str

        :param destination_account_type: Type of the destination account (optional)
        :type destination_account_type: str

        :param document_type: Document type (optional)
        :type document_type: str

        :param document_number: Document number (optional)
        :type document_number: str

        :param country: Country (optional)
        :type country: str

        :param destination_bank_code: Bank code of the destination (optional)
        :type destination_bank_code: str

        :param payment_intent_id: Payment intent ID (optional)
        :type payment_intent_id: str

        :param external_id: External ID (optional)
        :type external_id: str

        :param mobile_os: Mobile OS (optional)
        :type mobile_os: str

        :rtype: :class:`~prometeo.banking.models.PreprocessTransfer`
        """
        data = await self._client.preprocess_transfer(
            self._session_key,
            origin_account,
            destination_institution,
            destination_account,
            currency,
            amount,
            concept,
            destination_owner_name,
            branch,
        )
        return PreprocessTransfer(**data["result"])

    @utils.adapt_async_sync
    async def confirm_transfer(
        self,
        request_id,
        authorization_type,
        authorization_data,
        authorization_device_number=None,
    ):
        """
        Confirm transfer.

        :param request_id: Id of the request returned by the endpoint of
                           :meth:`~prometeo.banking.client.BankingAPIClient.preprocess_transfer`
        :type request_id: str

        :param authorization_type:
            - ``cardCode`` Coordinates card
            - ``pin`` Account personal pin
            -
                ``otp`` One-time generated pin, sent by sms, email,
                hard token or soft token.
            - ``otp-api`` Hard token or soft token device, digitized by Promete

        :type authorization_type: str

        :param authorization_data: Verification value (pin number,
                                   coordinates card response , etc) if there
                                   are several values, they must be separated by commas.
        :type authorization_data: str

        :param authorization_device_number: OTP Device Serial Number
        :type authorization_device_number: str

        :rtype: :class:`~prometeo.banking.models.ConfirmTransfer`
        """
        data = await self._client.confirm_transfer(
            self._session_key,
            request_id,
            authorization_type,
            authorization_data,
            authorization_device_number,
        )
        return ConfirmTransfer(**data["transfer"])

    @utils.adapt_async_sync
    async def list_transfer_institutions(self):
        """
        List transfer institutions.

        :rtype: :class:`~prometeo.banking.models.TransferInstitution`
        """
        data = await self._client.list_transfer_institutions(self._session_key)
        return [
            TransferInstitution(**institution) for institution in data["destinations"]
        ]


class BankingAPIClient(base_client.BaseClient):
    """
    API Client for banking api
    """

    ENVIRONMENTS = {
        "production": PRODUCTION_URL,
        "sandbox": SANDBOX_URL,
    }

    session_class = Session

    def on_response(self, data):
        if data["status"] == "error":
            if data["message"] == "Invalid key":
                raise exceptions.InvalidSessionKeyError(data["message"])
            else:
                raise BankingClientError(data["message"])

    @utils.adapt_async_sync
    async def login(self, provider, username, password, session_key=None, **kwargs):
        headers = {"X-Session-Key": session_key} if session_key else {}
        return await self.call_api(
            "POST",
            "/login/",
            data={
                "provider": provider,
                "username": username,
                "password": password,
                **kwargs,
            },
            headers=headers,
        )

    @utils.adapt_async_sync
    async def login_procedure(self, session_key, **kwargs):
        return await self.call_api(
            "POST",
            "/login-procedure/",
            data={
                **kwargs,
            },
            headers={"X-Session-Key": session_key},
        )

    @utils.adapt_async_sync
    async def get_owner_info(self, session_key):
        return await self.call_api(
            "GET",
            "/info/",
            headers={"X-Session-Key": session_key},
        )

    @utils.adapt_async_sync
    async def get_transfer_authorization_methods(self, session_key):
        return await self.call_api(
            "GET",
            "/transfer/mfa-methods",
            headers={"X-Session-Key": session_key},
        )

    @utils.adapt_async_sync
    async def get_clients(self, session_key):
        return await self.call_api(
            "GET",
            "/client/",
            headers={"X-Session-Key": session_key},
        )

    @utils.adapt_async_sync
    async def select_client(self, session_key, client_id):
        return await self.call_api(
            "GET",
            "/client/{}/".format(client_id),
            headers={"X-Session-Key": session_key},
        )

    @utils.adapt_async_sync
    async def get_accounts(self, session_key):
        return await self.call_api(
            "GET",
            "/account/",
            headers={"X-Session-Key": session_key},
        )

    @utils.adapt_async_sync
    async def get_movements(
        self, session_key, account_number, currency_code, date_start, date_end
    ):
        return await self.call_api(
            "GET",
            "/movement/",
            headers={"X-Session-Key": session_key},
            params={
                "account": account_number,
                "currency": currency_code,
                "date_start": date_start.strftime("%d/%m/%Y"),
                "date_end": date_end.strftime("%d/%m/%Y"),
            },
        )

    @utils.adapt_async_sync
    async def get_credit_cards(self, session_key):
        return await self.call_api(
            "GET",
            "/credit-card/",
            headers={"X-Session-Key": session_key},
        )

    @utils.adapt_async_sync
    async def get_credit_card_movements(
        self, session_key, card_number, currency_code, date_start, date_end
    ):
        return await self.call_api(
            "GET",
            f"/credit-card/{card_number}/movements",
            headers={"X-Session-Key": session_key},
            params={
                "currency": currency_code,
                "date_start": date_start.strftime("%d/%m/%Y"),
                "date_end": date_end.strftime("%d/%m/%Y"),
            },
        )

    @utils.adapt_async_sync
    async def get_providers(self):
        return await self.call_api("GET", "/provider/")

    @utils.adapt_async_sync
    async def get_provider_detail(self, provider_code, key=None, value=None):
        params = {"key": key, "value": value} if key and value else None
        return await self.call_api("GET", f"/provider/{provider_code}/", params=params)

    @utils.adapt_async_sync
    async def logout(self, session_key):
        return await self.call_api(
            "GET",
            "/logout/",
            headers={"X-Session-Key": session_key},
        )

    @utils.adapt_async_sync
    async def preprocess_transfer(
        self,
        session_key,
        origin_account,
        destination_institution,
        destination_account,
        currency,
        amount,
        concept,
        destination_owner_name,
        branch,
        destination_account_type=None,
        document_type=None,
        document_number=None,
        country=None,
        destination_bank_code=None,
        payment_intent_id="",
        external_id=None,
        mobile_os=None,
        **kwargs,
    ):
        return await self.call_api(
            "POST",
            "/transfer/preprocess",
            headers={
                "X-Session-Key": session_key,
                "Payment-Intent-ID": payment_intent_id,
            },
            data={
                "origin_account": origin_account,
                "destination_institution": destination_institution,
                "destination_account": destination_account,
                "currency": currency,
                "amount": amount,
                "concept": concept,
                "destination_owner_name": destination_owner_name,
                "branch": branch,
                "destination_account_type": destination_account_type,
                "document_type": document_type,
                "document_number": document_number,
                "country": country,
                "destination_bank_code": destination_bank_code,
                "mobile_os": mobile_os,
                "external_id": external_id,
                **kwargs,
            },
        )

    @utils.adapt_async_sync
    async def retry_preprocess_transfer(
        self,
        session_key,
        request_id,
        **kwargs,
    ):
        return await self.call_api(
            "POST",
            "/transfer/preprocess/retry",
            headers={
                "X-Session-Key": session_key,
            },
            data={
                "request_id": request_id,
                **kwargs,
            },
        )

    @utils.adapt_async_sync
    async def confirm_transfer(
        self,
        session_key,
        request_id,
        authorization_type,
        authorization_data,
        authorization_device_number=None,
    ):
        return await self.call_api(
            "POST",
            "/transfer/confirm",
            headers={"X-Session-Key": session_key},
            data={
                "request_id": request_id,
                "authorization_type": authorization_type,
                "authorization_data": authorization_data,
                "authorization_device_number": authorization_device_number,
            },
        )

    @utils.adapt_async_sync
    async def list_transfer_institutions(self, session_key):
        return await self.call_api(
            "GET",
            "/transfer/destinations",
            headers={"X-Session-Key": session_key},
        )


class Account(object):
    """
    A bank account, returned by :meth:`~prometeo.banking.client.Session.get_accounts`
    """

    def __init__(self, client, session_key, account_data):
        self._client = client
        self._session_key = session_key

        self.id = account_data.id
        self.name = account_data.name
        self.number = account_data.number
        self.branch = account_data.branch
        self.currency = account_data.currency
        self.balance = account_data.balance

    @utils.adapt_async_sync
    async def get_movements(self, date_start, date_end):
        """
        List an account's movements for a range of dates.

        :param date_start: Start of the date range for movements.
        :type date_start: :class:`~datetime.datetime`

        :param date_end: End of the date range for movements.
        :type date_end: :class:`~datetime.datetime`

        :rtype: List of :class:`~prometeo.banking.models.Movement`
        """
        data = await self._client.get_movements(
            self._session_key,
            self.number,
            self.currency,
            date_start,
            date_end,
        )
        return [
            Movement(
                id=movement["id"],
                reference=movement["reference"],
                date=datetime.strptime(movement["date"], "%d/%m/%Y"),
                detail=movement["detail"],
                debit=movement["debit"],
                credit=movement["credit"],
                extra_data=movement["extra_data"],
            )
            for movement in data["movements"]
        ]


class CreditCard(object):
    """
    A credit card, returned by :meth:`~prometeo.banking.client.Session.get_credit_cards`
    """

    def __init__(self, client, session_key, card_data):
        self._client = client
        self._session_key = session_key

        self.id = card_data.id
        self.name = card_data.name
        self.number = card_data.number
        self.close_data = card_data.close_date
        self.due_date = card_data.due_date
        self.balance_local = card_data.balance_local
        self.balance_dollar = card_data.balance_dollar

    @utils.adapt_async_sync
    async def get_movements(self, currency_code, date_start, date_end):
        """
        List credit card's movements for a range of dates.

        :param date_start: Start of the date range for movements.
        :type date_start: :class:`~datetime.datetime`

        :param date_end: End of the date range for movements.
        :type date_end: :class:`~datetime.datetime`

        :rtype: List of :class:`~prometeo.banking.models.Movement`
        """
        data = await self._client.get_credit_card_movements(
            self._session_key,
            self.number,
            currency_code,
            date_start,
            date_end,
        )
        return [
            Movement(
                id=movement["id"],
                reference=movement["reference"],
                date=datetime.strptime(movement["date"], "%d/%m/%Y"),
                detail=movement["detail"],
                debit=movement["debit"],
                credit=movement["credit"],
                extra_data=movement.get("extra_data"),
            )
            for movement in data["movements"]
        ]



from typing import List, Optional
from pydantic import BaseModel


class CreatePaymentIntentResponse(BaseModel):
    intent_id: str
    external_id: Optional[str]
    concept: str
    currency: str
    amount: str
    email: Optional[str]
    bank_codes: List[str]


class StatusHistory(BaseModel):
    status: str
    message: Optional[str]
    error_type: Optional[str]
    error_code: Optional[str]
    provider_code: Optional[str]
    timestamp: str


class Customer(BaseModel):
    name: str
    document_type: Optional[str]
    document_number: Optional[str]
    email: str


class PaymentIntent(BaseModel):
    intent_id: str
    product_id: str
    external_id: Optional[str]
    concept: str
    currency: str
    amount: float
    status_history: List[StatusHistory]
    customer: Optional[Customer]
    current_status: str



from prometeo import exceptions
from prometeo.payment import exceptions as payment_exceptions
from tests.base_test_case import BaseTestCase

from prometeo import Client
import respx


class TestClient(BaseTestCase):
    def setUp(self):
        self.client = Client("test_key", "production")

    @respx.mock
    def test_create_intent_error_concept_too_long(self):
        self.mock_post_request(
            respx,
            "/api/v1/payment-intent/",
            json={
                "concept": ["Ensure this value has at most 20 characters (it has 32)."]
            },
            status_code=400,
        )
        with self.assertRaises(
            payment_exceptions.PaymentInvalidParameterClientError
        ) as e:
            self.client.payment.create_intent(
                widget_id="BBBBB-d66a-4234-8ea1-3e3ce1c0935b",
                currency="USD",
                amount="1.00",
                external_id=None,
                concept="TOO LONG CONCEPT MAX IS 20 CHARS",
            )

        self.assertIn("concept", e.exception.param)
        self.assertIn(
            "Ensure this value has at most 20 characters (it has 32).",
            e.exception.message,
        )

    @respx.mock
    def test_create_intent_error_inexistent_bank_code(self):
        self.mock_post_request(
            respx,
            "/api/v1/payment-intent/",
            json={
                "bank_codes": ["Object with code=inexistent_bank_code does not exist."]
            },
            status_code=400,
        )
        with self.assertRaises(
            payment_exceptions.PaymentInvalidParameterClientError
        ) as e:
            self.client.payment.create_intent(
                widget_id="BBBBB-d66a-4234-8ea1-3e3ce1c0935b",
                currency="USD",
                amount="1.00",
                external_id=None,
                concept="PROM123452 REF454243",
                bank_codes=["inexistent_bank"],
            )

        self.assertIn("bank_codes", e.exception.param)
        self.assertIn(
            "Object with code=inexistent_bank_code does not exist.", e.exception.message
        )

    @respx.mock
    def test_create_intent_successfully(self):
        self.mock_post_request(
            respx, "/api/v1/payment-intent/", "create_intent_response_success"
        )
        data = self.client.payment.create_intent(
            widget_id="BBBBB-d66a-4234-8ea1-3e3ce1c0935b",
            currency="USD",
            amount="1.00",
            external_id=None,
            concept="PROM123452 REF454243",
            bank_codes=["test"],
        )
        self.assertEqual(data.intent_id, "52abd9b2-7a8d-4e8b-b724-0b68701d2e0e")

    @respx.mock
    def test_valid_intent_id_unpaid(self):
        intent_id = "bea71e55-a1ec-4e5f-a5c0-c0e10b1a571c"
        self.mock_get_request(
            respx, f"/api/v1/payment-intent/{intent_id}", "valid_intent_unpaid"
        )

        data = self.client.payment.get_transaction_data(intent_id)
        self.assertEqual(1118.86, data.amount)
        self.assertEqual("PEN", data.currency)

    @respx.mock
    def test_valid_intent_id_paid(self):
        intent_id = "9999e375-9999-4da0-a4fd-85e5a04c9999"
        self.mock_get_request(
            respx, f"/api/v1/payment-intent/{intent_id}", "valid_intent_paid"
        )

        data = self.client.payment.get_transaction_data(intent_id)
        self.assertEqual(133.89, data.amount)
        self.assertEqual("UYU", data.currency)

    @respx.mock
    def test_invalid_intent_id(self):
        intent_id = "bea71e55-a1ec-4e5f-a5c0-c0e10b1a571c"
        self.mock_get_request(
            respx,
            f"/api/v1/payment-intent/{intent_id}",
            json={"message": "The payment intent id is not valid"},
            status_code=400,
        )
        with self.assertRaises(exceptions.BadRequestError):
            self.client.payment.get_transaction_data(intent_id)



from datetime import datetime
from typing import List, Optional, Union
from pydantic import BaseModel


class Client(BaseModel):
    id: str
    name: str


class Account(BaseModel):
    id: str
    name: str
    number: str
    branch: str
    currency: str
    balance: float


class CreditCard(BaseModel):
    id: Union[str, int]
    name: str
    number: str
    close_date: Union[datetime, str]
    due_date: Union[datetime, str]
    balance_local: Union[float, str]
    balance_dollar: Union[float, str]


class Movement(BaseModel):
    id: Union[str, int]
    reference: str
    date: Union[datetime, str]
    detail: str
    debit: Optional[Union[float, str]]
    credit: Optional[Union[float, str]]
    extra_data: Optional[dict]


class Provider(BaseModel):
    code: str
    country: str
    name: str


class OTP(BaseModel):
    data: Optional[List[str]]
    type: str


class PreprocessTransfer(BaseModel):
    approved: bool
    authorization_devices: Optional[List[OTP]]
    message: Optional[str]
    request_id: str


class ConfirmTransfer(BaseModel):
    message: Optional[str]
    success: bool


class TransferInstitution(BaseModel):
    id: int
    name: str


class AuthFieldChoice(BaseModel):
    name: str
    type: str
    interactive: bool
    optional: bool
    label_es: str
    label_en: str
    choices: List[dict]
    default: Optional[str]


class AuthFieldText(BaseModel):
    name: str
    type: str
    interactive: bool
    optional: bool
    label_es: str
    label_en: str


class AuthFieldPassword(BaseModel):
    name: str
    type: str
    interactive: bool
    optional: bool
    label_es: str
    label_en: str


class EndpointStatus(BaseModel):
    endpoint: str
    status: str
    timestamp: str


class EndpointsStatus(BaseModel):
    test: List[EndpointStatus]
    prod: List[EndpointStatus]


class AccountType(BaseModel):
    name: str
    label_es: str
    label_en: str


class Bank(BaseModel):
    code: str
    name: str
    logo: str


class Methods(BaseModel):
    accounts: bool
    credit_cards: bool
    accounts_movements: bool
    credit_cards_movements: bool
    personal_info: bool
    transfers: bool
    enrollments: bool


class ProviderDetail(BaseModel):
    name: str
    aliases: List[str]
    country: str
    auth_fields: List[Union[AuthFieldChoice, AuthFieldText, AuthFieldPassword]]
    endpoints_status: Optional[EndpointsStatus]
    account_type: List[AccountType]
    logo: str
    bank: Bank
    methods: Optional[Methods]



from datetime import datetime
from six.moves.urllib.parse import parse_qs


from prometeo.sat.client import (
    LoginScope,
    BillStatus,
    DownloadAction,
    Motive,
    DocumentType,
    Status,
    SendType,
)
from tests.base_test_case import BaseTestCase
import respx


class TestClient(BaseTestCase):
    def setUp(self):
        super(TestClient, self).setUp()
        self.session_key = "123456"

    @respx.mock
    async def test_login_success(self):
        self.mock_post_request(respx, "/login/", "successful_login")
        rfc = "ABC12345DEF"
        password = "test_password"
        session = await self.client.sat.login(rfc, password, LoginScope.CFDI)
        self.assertEqual("logged_in", session.get_status())
        self.assertEqual(self.session_key, session.get_session_key())

        last_request = respx.calls.last.request
        request_body = parse_qs(str(last_request.content.decode("utf-8")))
        self.assertEqual("sat", request_body["provider"][0])
        self.assertEqual(rfc, request_body["rfc"][0])
        self.assertEqual(password, request_body["password"][0])
        self.assertEqual("cfdi", request_body["scope"][0])

    @respx.mock
    async def test_logout(self):
        self.mock_get_request(respx, "/logout/", "successful_logout")
        await self.client.sat.logout(self.session_key)
        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])

    @respx.mock
    async def test_emitted(self):
        self.mock_get_request(respx, "/cfdi/emitted/", "cfdi_emitted_list")
        await self.client.sat.get_emitted(
            self.session_key,
            datetime(2019, 1, 1),
            datetime(2019, 5, 1),
            BillStatus.ANY,
            DownloadAction.LIST,
        )

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual("01/01/2019", self.qs(last_request)["date_start"][0])
        self.assertEqual("01/05/2019", self.qs(last_request)["date_end"][0])
        self.assertEqual("any", self.qs(last_request)["status"][0])
        self.assertEqual("list", self.qs(last_request)["action"][0])

    @respx.mock
    async def test_get_emitted_list(self):
        self.mock_get_request(respx, "/cfdi/emitted/", "cfdi_emitted_list")
        emitted = await self.client.sat.get_emitted(
            self.session_key,
            datetime(2019, 1, 1),
            datetime(2019, 5, 1),
            BillStatus.ANY,
            DownloadAction.LIST,
        )

        self.assertEqual(3, len(emitted))
        self.assertEqual(
            datetime(2018, 8, 29, 20, 50, 4), emitted[0].certification_date
        )
        self.assertEqual(datetime(2018, 8, 29, 20, 50, 3), emitted[0].emitted_date)
        self.assertEqual(BillStatus.VALID.value, emitted[0].status)

    @respx.mock
    async def test_get_emitted_bulk_download(self):
        self.mock_get_request(respx, "/cfdi/emitted/", "cfdi_emitted_bulk_download")
        downloads = await self.client.sat.get_emitted(
            self.session_key,
            datetime(2019, 1, 1),
            datetime(2019, 5, 1),
            BillStatus.ANY,
            DownloadAction.BULK_DOWNLOAD,
        )
        last_request = respx.calls.last.request
        self.assertEqual("bulk_download", self.qs(last_request)["action"][0])
        self.assertEqual(2, len(downloads))
        self.assertEqual(
            "50AD2BA1-27AE-4CC3-84FD-265E585A1F67", downloads[0].request_id
        )

    @respx.mock
    async def test_get_emitted_metadata_download(self):
        self.mock_get_request(respx, "/cfdi/emitted/", "cfdi_emitted_bulk_download")
        downloads = await self.client.sat.get_emitted(
            self.session_key,
            datetime(2019, 1, 1),
            datetime(2019, 5, 1),
            BillStatus.ANY,
            DownloadAction.METADATA_DOWNLOAD,
        )

        last_request = respx.calls.last.request
        self.assertEqual("metadata_download", self.qs(last_request)["action"][0])
        self.assertEqual(2, len(downloads))
        self.assertEqual(
            "50AD2BA1-27AE-4CC3-84FD-265E585A1F67", downloads[0].request_id
        )

    @respx.mock
    async def test_get_emitted_pdf_export(self):
        self.mock_get_request(respx, "/cfdi/emitted/", "cfdi_emitted_pdf_export")
        downloads = await self.client.sat.get_emitted(
            self.session_key,
            datetime(2019, 1, 1),
            datetime(2019, 5, 1),
            BillStatus.ANY,
            DownloadAction.PDF_EXPORT,
        )

        last_request = respx.calls.last.request
        self.assertEqual("pdf_export", self.qs(last_request)["action"][0])
        self.assertEqual(1, len(downloads))
        self.assertEqual(
            "/pdf/655ea3b7d0ec5fa1914aca7809e72449.pdf", downloads[0].pdf_url
        )

    @respx.mock
    async def test_download_emitted(self):
        bill_id = "DDAA8B0B-4FDC-43D7-A633-F307B898AB3C"
        self.mock_get_request(
            respx, "/cfdi/emitted/{}/".format(bill_id), "cfdi_emitted_download"
        )
        download = await self.client.sat.download_emitted(self.session_key, bill_id)

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual(
            download.download_url, "/download/4f3882b1d413f761ced91b6bd583f6ee.xml"
        )

    @respx.mock
    async def test_get_received(self):
        self.mock_get_request(respx, "/cfdi/received/", "cfdi_received_list")
        await self.client.sat.get_received(
            self.session_key,
            2019,
            5,
            BillStatus.ANY,
            DownloadAction.LIST,
        )

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual("2019", self.qs(last_request)["year"][0])
        self.assertEqual("5", self.qs(last_request)["month"][0])
        self.assertEqual("any", self.qs(last_request)["status"][0])
        self.assertEqual("list", self.qs(last_request)["action"][0])

    @respx.mock
    async def test_get_received_list(self):
        self.mock_get_request(respx, "/cfdi/received/", "cfdi_received_list")
        received = await self.client.sat.get_received(
            self.session_key,
            2019,
            5,
            BillStatus.ANY,
            DownloadAction.LIST,
        )

        self.assertEqual(3, len(received))
        self.assertEqual(
            datetime(2018, 8, 29, 20, 50, 4), received[0].certification_date
        )
        self.assertEqual(datetime(2018, 8, 29, 20, 50, 3), received[0].emitted_date)
        self.assertEqual(BillStatus.VALID.value, received[0].status)

    @respx.mock
    async def test_get_received_bulk_download(self):
        self.mock_get_request(respx, "/cfdi/received/", "cfdi_received_bulk_download")
        downloads = await self.client.sat.get_received(
            self.session_key,
            2019,
            5,
            BillStatus.ANY,
            DownloadAction.BULK_DOWNLOAD,
        )
        last_request = respx.calls.last.request
        self.assertEqual("bulk_download", self.qs(last_request)["action"][0])
        self.assertEqual(2, len(downloads))
        self.assertEqual(
            "50AD2BA1-27AE-4CC3-84FD-265E585A1F67", downloads[0].request_id
        )

    @respx.mock
    async def test_get_received_metadata_download(self):
        self.mock_get_request(respx, "/cfdi/received/", "cfdi_received_bulk_download")
        downloads = await self.client.sat.get_received(
            self.session_key,
            2019,
            5,
            BillStatus.ANY,
            DownloadAction.METADATA_DOWNLOAD,
        )

        last_request = respx.calls.last.request
        self.assertEqual("metadata_download", self.qs(last_request)["action"][0])
        self.assertEqual(2, len(downloads))
        self.assertEqual(
            "50AD2BA1-27AE-4CC3-84FD-265E585A1F67", downloads[0].request_id
        )

    @respx.mock
    async def test_get_received_pdf_export(self):
        self.mock_get_request(respx, "/cfdi/received/", "cfdi_received_pdf_export")
        downloads = await self.client.sat.get_received(
            self.session_key,
            2019,
            5,
            BillStatus.ANY,
            DownloadAction.PDF_EXPORT,
        )
        last_request = respx.calls.last.request
        self.assertEqual("pdf_export", self.qs(last_request)["action"][0])
        self.assertEqual(1, len(downloads))
        self.assertEqual(
            "/pdf/655ea3b7d0ec5fa1914aca7809e72449.pdf", downloads[0].pdf_url
        )

    @respx.mock
    async def test_download_received(self):
        bill_id = "DDAA8B0B-4FDC-43D7-A633-F307B898AB3C"
        self.mock_get_request(
            respx, "/cfdi/received/{}/".format(bill_id), "cfdi_received_download"
        )
        download = await self.client.sat.download_received(self.session_key, bill_id)

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual(
            download.download_url, "/download/4f3882b1d413f761ced91b6bd583f6ee.xml"
        )

    @respx.mock
    async def test_get_downloads(self):
        self.mock_get_request(respx, "/cfdi/download/", "cfdi_list_downloads")
        downloads = await self.client.sat.get_downloads(self.session_key)
        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual(2, len(downloads))
        self.assertEqual(
            "CAF80916-71AC-4CEA-BECF-F2AC3F394948", downloads[0].request_id
        )

    @respx.mock
    async def test_get_download(self):
        request_id = "DDAA8B0B-4FDC-43D7-A633-F307B898AB3C"
        self.mock_get_request(
            respx, "/cfdi/download/{}/".format(request_id), "cfdi_download"
        )
        download = await self.client.sat.get_download(self.session_key, request_id)
        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual(
            download.download_url, "/download/4f3882b1d413f761ced91b6bd583f6ee.zip"
        )

    @respx.mock
    async def test_get_acknowledgements(self):
        self.mock_get_request(
            respx, "/ccee/acknowledgment/", "ccee_list_acknowledgements"
        )
        acks = await self.client.sat.get_acknowledgements(
            self.session_key,
            2018,
            1,
            5,
            Motive.AF,
            DocumentType.CT,
            Status.RECEIVED,
            SendType.N,
        )

        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual("2018", self.qs(last_request)["year"][0])
        self.assertEqual("1", self.qs(last_request)["month_start"][0])
        self.assertEqual("5", self.qs(last_request)["month_end"][0])
        self.assertEqual("af", self.qs(last_request)["motive"][0])
        self.assertEqual("ct", self.qs(last_request)["document_type"][0])
        self.assertEqual("received", self.qs(last_request)["status"][0])
        self.assertEqual("n", self.qs(last_request)["send_type"][0])
        self.assertEqual(3, len(acks))
        self.assertEqual("0002180100000000325368", acks[0].id)

    @respx.mock
    async def test_download_acknowldegement(self):
        ack_id = "0002180100000000325368"
        self.mock_get_request(
            respx,
            "/ccee/acknowledgment/{}/".format(ack_id),
            "ccee_download_acknowledgement",
        )
        download = await self.client.sat.download_acknowledgement(
            self.session_key, ack_id
        )
        last_request = respx.calls.last.request
        self.assertEqual(self.session_key, self.qs(last_request)["session_key"][0])
        self.assertEqual(
            "/download/4f3882b1d413f761ced91b6bd583f6ee.zip", download.download_url
        )



from prometeo import exceptions


class AccountValidationClientError(exceptions.PrometeoError):
    """Base exception class for errors in the AccountValidationClient."""

    pass


class InvalidAccountError(AccountValidationClientError):
    """Exception for invalid account error"""

    pass


class PendingValidationError(AccountValidationClientError):
    """Exception raised for pending validations"""

    pass


class CommunicationError(AccountValidationClientError):
    """Exception for communication-related errors."""

    pass


class MethodNotAvailableError(AccountValidationClientError):
    """Exception for method not available errors."""

    pass


class BankProviderNotAvailableError(AccountValidationClientError):
    """Exception for bank provider not available errors."""

    pass


class CountryNotAvailableError(AccountValidationClientError):
    """Exception for country not available errors."""

    pass


class InvalidParameterError(AccountValidationClientError):
    """Exception for invalid parameter errors."""

    pass



from datetime import datetime
from pydantic import BaseModel


class CFDIBill(BaseModel):
    id: str
    emitter_rfc: str
    emitter_reason: str
    receiver_rfc: str
    receiver_reason: str
    emitted_date: datetime
    certification_date: datetime
    certification_pac: str
    total_value: float
    effect: str
    status: str


class CFDIDownloadItem(BaseModel):
    request_id: str
    type: str
    count: int


class DownloadRequest(BaseModel):
    request_id: str


class PdfFile(BaseModel):
    pdf_url: str


class DownloadFile(BaseModel):
    download_url: str


class AcknowledgementResult(BaseModel):
    id: str
    period: str
    motive: str
    document_type: str
    send_type: str
    file_name: str
    reception_date: str
    status: str



from datetime import datetime


from prometeo.sat.client import (
    SatAPIClient,
    Session,
    BillStatus,
    Motive,
    DocumentType,
    Status,
    SendType,
)
from tests.base_test_case import BaseTestCase
import respx
from six.moves.urllib.parse import parse_qs, urlparse


class TestSession(BaseTestCase):
    def setUp(self):
        super(TestSession, self).setUp()
        client = SatAPIClient("test_api_key", "sandbox")
        self.session_key = "test_session_key"
        self.session = Session(client, "logged_in", self.session_key)

    @respx.mock
    async def test_logout(self):
        self.mock_get_request(respx, "/logout/", "successful_logout")
        await self.session.logout()
        last_request = respx.calls.last.request
        self.assertIn(self.session_key, str(last_request.url))

    @respx.mock
    def test_get_emitted_list(self):
        self.mock_get_request(respx, "/cfdi/emitted/", "cfdi_emitted_list")
        bills = self.session.get_emitted_bills(
            datetime(2019, 1, 1),
            datetime(2019, 5, 1),
            BillStatus.ANY,
        )

        self.assertEqual(3, len(bills))
        self.assertEqual(datetime(2018, 8, 29, 20, 50, 4), bills[0].certification_date)
        self.assertEqual(datetime(2018, 8, 29, 20, 50, 3), bills[0].emitted_date)
        self.assertEqual(BillStatus.VALID.value, bills[0].status)

    @respx.mock
    def test_download_emitted(self):
        self.mock_get_request(respx, "/cfdi/emitted/", "cfdi_emitted_bulk_download")
        downloads = self.session.download_emitted_bills(
            datetime(2019, 1, 1),
            datetime(2019, 5, 1),
            BillStatus.ANY,
        )
        last_request = respx.calls.last.request
        qs = parse_qs(urlparse(str(last_request.url)).query)
        self.assertEqual("bulk_download", qs["action"][0])
        self.assertEqual(2, len(downloads))
        self.assertEqual(
            "50AD2BA1-27AE-4CC3-84FD-265E585A1F67", downloads[0].request_id
        )

    @respx.mock
    def test_get_received_list(self):
        self.mock_get_request(respx, "/cfdi/received/", "cfdi_received_list")
        bills = self.session.get_received_bills(2018, 8, BillStatus.ANY)

        self.assertEqual(3, len(bills))
        self.assertEqual(datetime(2018, 8, 29, 20, 50, 4), bills[0].certification_date)
        self.assertEqual(datetime(2018, 8, 29, 20, 50, 3), bills[0].emitted_date)
        self.assertEqual(BillStatus.VALID.value, bills[0].status)

    @respx.mock
    def test_download_received(self):
        self.mock_get_request(respx, "/cfdi/received/", "cfdi_received_bulk_download")
        downloads = self.session.download_received_bills(2018, 5, BillStatus.ANY)
        last_request = respx.calls.last.request
        qs = parse_qs(urlparse(str(last_request.url)).query)
        self.assertEqual("bulk_download", qs["action"][0])
        self.assertEqual(2, len(downloads))
        self.assertEqual(
            "50AD2BA1-27AE-4CC3-84FD-265E585A1F67", downloads[0].request_id
        )

    @respx.mock
    async def test_get_downloads(self):
        self.mock_get_request(respx, "/cfdi/download/", "cfdi_list_downloads")
        downloads = await self.session.get_downloads()

        last_request = respx.calls.last.request
        self.assertIn(self.session_key, str(last_request.url))
        self.assertEqual(2, len(downloads))
        self.assertEqual(
            "CAF80916-71AC-4CEA-BECF-F2AC3F394948", downloads[0].request_id
        )

    @respx.mock
    async def test_get_download_from_list(self):
        request_id = "CAF80916-71AC-4CEA-BECF-F2AC3F394948"
        self.mock_get_request(respx, "/cfdi/download/", "cfdi_list_downloads")
        self.mock_get_request(
            respx, "/cfdi/download/{}/".format(request_id), "cfdi_download"
        )
        download_requests = await self.session.get_downloads()
        download = await download_requests[0].get_download()
        self.assertEqual(download.url, "/download/4f3882b1d413f761ced91b6bd583f6ee.zip")

    @respx.mock
    async def test_get_download(self):
        request_id = "50AD2BA1-27AE-4CC3-84FD-265E585A1F67"
        self.mock_get_request(respx, "/cfdi/received/", "cfdi_received_bulk_download")
        self.mock_get_request(
            respx, "/cfdi/download/{}/".format(request_id), "cfdi_download"
        )

        download_requests = await self.session.download_received_bills(
            2018, 5, BillStatus.ANY
        )
        download = await download_requests[0].get_download()
        self.assertEqual(download.url, "/download/4f3882b1d413f761ced91b6bd583f6ee.zip")

    @respx.mock
    def test_get_acknowledgements(self):
        self.mock_get_request(
            respx, "/ccee/acknowledgment/", "ccee_list_acknowledgements"
        )

        acks = self.session.get_acknowledgements(
            year=2018,
            month_start=1,
            month_end=5,
            motive=Motive.AF,
            document_type=DocumentType.CT,
            status=Status.RECEIVED,
            send_type=SendType.N,
        )
        last_request = respx.calls.last.request
        qs = parse_qs(urlparse(str(last_request.url)).query)
        self.assertEqual(self.session_key, qs["session_key"][0])
        self.assertEqual("2018", qs["year"][0])
        self.assertEqual("1", qs["month_start"][0])
        self.assertEqual("5", qs["month_end"][0])
        self.assertEqual("af", qs["motive"][0])
        self.assertEqual("ct", qs["document_type"][0])
        self.assertEqual("received", qs["status"][0])
        self.assertEqual("n", qs["send_type"][0])
        self.assertEqual(3, len(acks))
        self.assertEqual("0002180100000000325368", acks[0].id)

    @respx.mock
    async def test_download_acknowldegement(self):
        self.mock_get_request(
            respx, "/ccee/acknowledgment/", "ccee_list_acknowledgements"
        )
        self.mock_get_request(
            respx,
            "/ccee/acknowledgment/0002180100000000325368/",
            "ccee_download_acknowledgement",
        )

        acks = await self.session.get_acknowledgements(
            year=2018,
            month_start=1,
            month_end=5,
            motive=Motive.AF,
            document_type=DocumentType.CT,
            status=Status.RECEIVED,
            send_type=SendType.N,
        )
        download = await acks[0].get_download()
        self.assertEqual(download.url, "/download/4f3882b1d413f761ced91b6bd583f6ee.zip")

    @respx.mock
    async def test_download_not_ready(self):
        self.mock_get_request(respx, "/cfdi/received/", "cfdi_received_bulk_download")
        self.mock_get_request(
            respx,
            "/cfdi/download/50AD2BA1-27AE-4CC3-84FD-265E585A1F67/",
            "not_found",
            status_code=404,
        )

        downloads = await self.session.download_received_bills(2018, 5, BillStatus.ANY)
        self.assertFalse(await downloads[0].is_ready())

    @respx.mock
    async def test_download_ready(self):
        self.mock_get_request(respx, "/cfdi/received/", "cfdi_received_bulk_download")
        self.mock_get_request(
            respx,
            "/cfdi/download/50AD2BA1-27AE-4CC3-84FD-265E585A1F67/",
            "cfdi_download",
        )

        downloads = await self.session.download_received_bills(2018, 5, BillStatus.ANY)
        self.assertTrue(await downloads[0].is_ready())

    @respx.mock
    def test_restore_session(self):
        self.mock_get_request(respx, "/cfdi/download/", "cfdi_list_downloads")

        session_key = "test_restored_key"
        session = self.client.sat.get_session(session_key)

        session.get_downloads()
        last_request = respx.calls.last.request
        self.assertIn("session_key", str(last_request.url.query))



class BaseSession(object):
    """
    Base class that handles calling the API endpoints that use session keys.
    """

    def __init__(self, client, status, session_key):
        self._client = client
        self._status = status
        self._session_key = session_key

    def get_status(self):
        """
        Returns this session's status

        :rtype: str
        """
        return self._status

    def get_session_key(self):
        """
        Returns this session's session key

        :rtype: str
        """
        return self._session_key



from prometeo.sat.client import SatAPIClient
from prometeo.base_client import Download
from tests.base_test_case import BaseTestCase
import respx


class TestSession(BaseTestCase):
    def setUp(self):
        super(TestSession, self).setUp()
        self.client = SatAPIClient("test_api_key", "sandbox")

    @respx.mock
    async def test_download_file(self):
        file_url = "/download/file.txt"
        file_content = "content"
        self.mock_get_request(respx, file_url, text=file_content)
        download = Download(self.client, file_url)
        file = await download.get_file()
        self.assertEqual(file.decode(), file_content)



from typing import List, Optional
from pydantic import BaseModel
from datetime import datetime


class Balance(BaseModel):
    type: str
    amount: float
    value: float


class Name(BaseModel):
    first_surname: str
    second_surname: str
    first_name: str
    other_names: str


class Field(BaseModel):
    name: str
    number: int
    value: float


class NumerationRange(BaseModel):
    from_number: int
    to_number: int
    mode: str
    establishment: str
    prefix: str
    type: str


class Numeration(BaseModel):
    nit: Optional[str]
    dv: Optional[str]
    name: Optional[Name]
    reason: Optional[str]
    address: Optional[str]
    country: Optional[str]
    department: Optional[str]
    municipality: Optional[str]
    ranges: List[NumerationRange]
    pdf_url: Optional[str]
    pdf_available: Optional[bool]


class RentDeclaration(BaseModel):
    pdf_url: Optional[str]
    pdf: Optional[object]
    fields: List[Field]
    year: int
    form_number: int
    nit: str
    dv: str
    name: Name
    reason: str
    direction_code: str
    economic_activity: str
    correction_code: str
    previous_form: Optional[str]


class VATDeclaration(BaseModel):
    pdf_url: Optional[str]
    pdf: Optional[object]
    fields: List[Field]
    year: int
    period: str
    form_number: int
    nit: str
    dv: str
    name: Name
    reason: str
    direction_code: str
    correction_code: str
    previous_form: Optional[str]


class Retentions(BaseModel):
    pdf_url: Optional[str]
    pdf: Optional[object]
    fields: List[Field]
    year: int
    period: int
    form_number: int
    nit: str
    reason: str
    direction_code: str


class Representative(BaseModel):
    document: str
    document_type: str
    name: Name
    representation_type: str
    start_date: datetime


class Member(BaseModel):
    document_type: str
    document: str
    nationality: str
    name: Name
    start_date: datetime


class Location(BaseModel):
    country: str
    city: str
    phone1: str
    phone2: str
    department: str
    address: str
    email: str


class CapitalComposition(BaseModel):
    national: str
    national_private: str
    national_public: str
    foreign: str
    foreign_private: str
    foreign_public: str


class Accountant(BaseModel):
    document: str
    start_date: datetime
    name: str
    professional_card: str


class CompanyInfo(BaseModel):
    accountant: Accountant
    capital_composition: CapitalComposition
    reason: str
    pdf_url: Optional[str]
    pdf: Optional[object]
    location: Location
    name: str
    constitution_date: datetime
    representation: List[Representative]
    members: List[Member]

    class Config:
        validate_assignment = True



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    requirements = fh.readlines()

setuptools.setup(
    name="prometeo",
    version="2.0.2",
    author="Prometeo",
    author_email="dev@prometeoapi.com",
    description="Python client library for Prometeo",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/prometeoapi/prometeo-python",
    packages=setuptools.find_packages(exclude="tests"),
    package_data={"README": ["README.md"]},
    include_package_data=True,
    python_requires=">=3.8",
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from datetime import datetime

from prometeo.banking.client import BankingAPIClient, Account
from prometeo.banking.models import Account as AccountModel

from six.moves.urllib.parse import parse_qs, urlparse
import respx
from tests.base_test_case import BaseTestCase


class TestAccount(BaseTestCase):
    def setUp(self):
        client = BankingAPIClient("test_api_key", "sandbox")
        self.session_key = "test_session_key"
        account = AccountModel(
            id="12345",
            name="Cuenta total",
            number="001234567890",
            branch="02 - 18 De Julio",
            currency="UYU",
            balance=1234.95,
        )
        self.account = Account(client, self.session_key, account)

    @respx.mock
    def test_get_movements(self):
        self.mock_get_request(
            respx,
            "/movement/",
            json={
                "movements": [
                    {
                        "credit": "",
                        "date": "12/01/2019",
                        "debit": 3500,
                        "detail": "RETIRO EFECTIVO CAJERO AUTOMATICO J.C. ",
                        "id": "-890185180",
                        "reference": "000000005084",
                        "extra_data": None,
                    },
                    {
                        "credit": "",
                        "date": "05/07/2019",
                        "debit": 16000,
                        "detail": "RETIRO EFECTIVO CAJERO AUTOMATICO J.H Y ",
                        "id": 1024917397,
                        "reference": "000000002931",
                        "extra_data": None,
                    },
                ],
                "status": "success",
            },
        )
        date_start = datetime(2019, 1, 1)
        date_end = datetime(2019, 12, 1)
        movements = self.account.get_movements(date_start, date_end)
        last_request = respx.calls.last.request
        qs = parse_qs(urlparse(str(last_request.url)).query)
        self.assertEqual(self.session_key, last_request.headers["X-Session-Key"])
        self.assertEqual(self.account.number, qs["account"][0])
        self.assertEqual(self.account.currency, qs["currency"][0])
        self.assertEqual("01/01/2019", qs["date_start"][0])
        self.assertEqual("01/12/2019", qs["date_end"][0])
        self.assertEqual(2, len(movements))



from prometeo import exceptions, base_client, utils
from .exceptions import (
    InvalidAccountError,
    PendingValidationError,
    MethodNotAvailableError,
    CommunicationError,
    BankProviderNotAvailableError,
    CountryNotAvailableError,
)
from typing import Optional, List, Tuple
from .models import AccountData
import re


PRODUCTION_URL = "https://account-validation.prometeoapi.net"
SANDBOX_URL = "https://account-validation.sandbox.prometeoapi.com"


class AccountValidationAPIClient(base_client.BaseClient):
    """
    API Client for Account Validation API
    """

    ENVIRONMENTS = {
        "production": PRODUCTION_URL,
        "sandbox": SANDBOX_URL,
    }

    def extract_invalid_parameters(
        self, error_message
    ) -> Tuple[Optional[List[str]], str]:
        pattern = r"Invalid parameter: ([^.,]+(?:,\s[^.,]+)*)[.]?\s*([^.]*)"
        match = re.match(pattern, error_message)
        if match:
            parameters = match.group(1)
            message = match.group(2)
            return parameters.split(", "), message.strip()
        return None, None

    def on_error(self, response, data):
        errors = data.get("errors", {}) or {}
        error_message = errors.get("message")
        error_code = errors.get("code")

        if error_code == 400 and error_message:
            raise exceptions.InvalidParameterError(
                *self.extract_invalid_parameters(error_message)
            )
        elif error_code == 404:
            raise InvalidAccountError(error_message)
        elif error_code == 202:
            raise PendingValidationError(error_message)
        elif error_code == 500:
            raise exceptions.InternalAPIError(error_message, response.text)
        elif error_code == 503:
            raise CommunicationError(error_message)
        elif error_code == 512:
            raise MethodNotAvailableError(error_message)
        elif error_code == 513:
            raise BankProviderNotAvailableError(error_message)
        elif error_code == 514:
            raise CountryNotAvailableError(error_message)

    @utils.adapt_async_sync
    async def validate(
        self,
        account_number: str,
        country_code: str,
        bank_code: Optional[str] = None,
        document_number: Optional[str] = None,
        branch_code: Optional[str] = None,
        account_type: Optional[List[str]] = None,
    ) -> AccountData:
        """
        Validate bank account information.

        :param account_number: The account number to be validated.
        :type account_number: str

        :param country_code: The country code associated with the account.
        :type country_code: str

        :param bank_code: The bank code if available.
        :type bank_code: Optional[str]

        :param document_number: The document number associated with the account.
        :type document_number: Optional[str]

        :param branch_code: The branch code if available.
        :type branch_code: Optional[str]

        :param account_type: A list of account types to consider.
        :type account_type: Optional[List[str]]

        :return: An object containing validated account information.
        :rtype: AccountData
        :raises: Any exceptions raised during the validation process.
        """

        data = await self.call_api(
            "POST",
            "/validate-account/",
            data={
                "account_number": account_number,
                "country_code": country_code,
                "document_number": document_number,
                "branch_code": branch_code,
                "bank_code": bank_code,
                "account_type": account_type,
            },
        )
        return AccountData(**data.get("data"))
